/*
 * Copyright (c) 2017-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <sys/time.h>
#include <sys/resource.h>

#include <infiniband/ibutils/timespec_utils.h>
#include <infiniband/ibutils/output_stream_primitives.h>
#include <infiniband/ibdiag/ibdiag_csv_out.h>
#include <infiniband/ibdm/OutputControl.h>


void CSVOut::SetCommentPos()
{
    IBDIAG_ENTER;

    idx_tbl_comment_pos = (size_t)sout.tellp() + string("# INDEX_TABLE ").size();
    sout << "# INDEX_TABLE " << CommentString(0, 0) << endl;
    sout << endl << endl;
    cur_CSV_line += 3;

    IBDIAG_RETURN_VOID;
}


int CSVOut::DumpStart(const char *name)
{
    IBDIAG_ENTER;

    clock_gettime(CLOCK_REALTIME, &m_began_tm);
    getrusage(RUSAGE_SELF, &m_began_usage);

    static const std::string csv_prefix = "csv:";

    OutputControl::Properties properties(csv_prefix + name);

    cur_idx.name             = name;
    current_section_disabled = false;

    if(properties.is_valid())
        if(!properties.enabled()) {
            current_section_disabled = true;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);
        }

    sout << "START_" << name << endl;

    cur_idx.offset = (size_t)sout.tellp();
    cur_idx.line   = ++cur_CSV_line;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void CSVOut::WriteBuf(const string &buf)
{
    IBDIAG_ENTER;

    if (current_section_disabled)
        IBDIAG_RETURN_VOID;

    const char* x = buf.c_str();

    while(*x)
    {
        if (*x == '\n')
            ++cur_CSV_line;
        else
            if ((*x < ' ' || *x > '~') && *x != '\t' && *x != '\r')
            {
                sout << std::hex << std::setfill('0') << std::setw(2)
                     << "\\u00" << +*x;
                ++x;
                continue;
            }

        sout << *x;
        ++x;
    }

    IBDIAG_RETURN_VOID;
}


void CSVOut::DumpEnd(const char *name)
{
    IBDIAG_ENTER;

    if(!current_section_disabled) {
        cur_idx.size = (size_t)sout.tellp() - cur_idx.offset;
        cur_idx.rows = cur_CSV_line - cur_idx.line - 1;
        index_table.push_back(cur_idx);

        sout << "END_" << cur_idx.name << endl;
        sout << endl << endl;
        cur_CSV_line += 3;
    }

    struct timespec real;
    struct timeval  user;
    struct timeval  sys;

    struct timespec end_tm;
    struct rusage   end_usage;

    clock_gettime(CLOCK_REALTIME, &end_tm);
    getrusage(RUSAGE_SELF, &end_usage);

    diff(end_tm, m_began_tm, real);
    diff(end_usage.ru_utime, m_began_usage.ru_utime, user);
    diff(end_usage.ru_stime, m_began_usage.ru_stime, sys);

    m_perf_info << cur_idx.name << ',' << !current_section_disabled << std::setfill('0') << ','
                << real.tv_sec << '.' << std::setw(6) << (real.tv_nsec / 1000) << ','
                << user.tv_sec << '.' << std::setw(6) << user.tv_usec << ','
                << sys.tv_sec << '.' << std::setw(6) << sys.tv_usec << '\n';

    current_section_disabled = true;
    IBDIAG_RETURN_VOID;
}

void CSVOut::Init()
{
    current_section_disabled = false,
    cur_CSV_line = 1,
    idx_tbl_comment_pos = 0;
    index_table.clear();

    cur_idx.name = "";
    cur_idx.offset = 0;
    cur_idx.size = 0;
    cur_idx.line = 0;
    cur_idx.rows = 0;
}


void CSVOut::DumpPerfTableCSV()
{
    IBDIAG_ENTER;

    if (!DumpStart(SECTION_CSV_PERF_INFO)) {
        WriteBuf(m_perf_info.str());
        DumpEnd(SECTION_CSV_PERF_INFO);
    }

    IBDIAG_RETURN_VOID;
}

void CSVOut::DumpIndexTableCSV()
{
    IBDIAG_ENTER;

    char buffer[256];

    size_t cur_offset = sout.tellp();

    sout << "START_" << SECTION_INDEX_TABLE << endl;
    sout << "Name," << " Offset," << " Size," << " Line," << " Rows" << endl;

    for (list_index_line::iterator it = index_table.begin(); it != index_table.end(); ++it) {
        snprintf(buffer, sizeof(buffer), "%s, %11ld, %11ld, %11ld, %11ld\n", it->name.c_str(), it->offset,
                it->size, it->line, it->rows);

        sout << buffer;
    }

    sout << "END_" << SECTION_INDEX_TABLE << endl;

    sout.seekp(idx_tbl_comment_pos);
    sout << CommentString(cur_offset, cur_CSV_line);

    IBDIAG_RETURN_VOID;
}


int CSVOut::Open(const char *file_name, string &err_message, bool is_custom, bool add_header)
{
    if (sout.is_open())
        return 1;

    Init();

    filename = file_name;

    int rc = IBFabric::OpenFile(OutputControl::Identity (file_name,
                                 (is_custom ? OutputControl::OutputControl_Flag_CustomFileName :
                                              OutputControl::OutputControl_Flag_None)),
                                 sout, filename, false, err_message, add_header);

    if(!rc)
        SetCommentPos();

    return rc;
}

void CSVOut::Close() {
    if (sout.is_open())
    {
        DumpPerfTableCSV();
        DumpIndexTableCSV();
        sout.close();
    }
}

