/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "infiniband/ibdm/Fabric.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibutils/output_stream_primitives.h"
#include <ctime>

#include "ibdiag_sim_info_dump.h"

const std::vector<const char*> SimInfoData::SMPCapMaskBits = {
    "IsPrivateLinearForwardingSupported",                  //   0
    "IsAdaptiveRoutingSupported",                          //   1
    "IsAdaptiveRoutingRev1Supported",                      //   2
    "IsRemotePortMirroringSupported",                      //   3
    "IsTemperatureSensingSupported",                       //   4
    "IsConfigSpaceAccessSupported",                        //   5
    "IsCableInfoSupported",                                //   6
    "IsSMPEyeOpenSupported",                               //   7

    "IsLossyVLConfigSupported",                            //   8
    "IsExtendedPortInfoSupported",                         //   9
    "IsAccessRegisterSupported",                           //  10
    "IsInterProcessCommunicationSupported",                //  11
    "IsPortSLToPrivateLFTMapSupported",                    //  12
    "IsExtendedNodeInfoSupported",                         //  13
    "IsSlaveVirtualSwitch",                                //  14
    "IsVirtualizationSupported",                           //  15

    "IsAdvancedRetransAndFECModesSupported",               //  16
    "IsSemaphoreLockSupported",                            //  17
    "IsSHArPAggrNodeSupported",                            //  18
    "IsCableInfoPasswordSupported",                        //  19
    "IsVL2VLPerSWIDSupported",                             //  20
    "IsSpecialPortsMarkingSupported",                      //  21
    "IsVPortGUIDInfoSupported",                            //  22
    "IsForcePLFTSubgroup0Supported",                       //  23

    "IsQoSConfigSLRateLimitSupported",                     //  24
    "IsQoSConfigSLVPortRateLimitSupported",                //  25
    "IsQoSConfigSLAllocBWSupported",                       //  26
    "IsQoSConfigSLVPortAllocBWSupported",                  //  27
    "IsEnhancedTrap128Supported",                          //  28
    "IsDevInfoSupported",                                  //  29
    "IsGlobalOOOSupported",                                //  30
    "IsMADAddlStatusSupported",                            //  31

    "IsQoSConfigSLRateLimitPortSupported",                 //  32
    "IsQoSConfigSLRateLimitVPortSupported",                //  33
    "IsQoSConfigSLAllocBWPortSupported",                   //  34
    "IsQoSConfigSLAllocBWVPortSupported",                  //  35
    "IsQoSConfigHighPriorityVLGroupRateLimitSupported",    //  36
    "IsQoSConfigHighPriorityVLGroupAllocBWSupported",      //  37
    "IsQoSConfigLowPriorityVLGroupRateLimitSupported",     //  38
    "IsQoSConfigLowPriorityVLGroupAllocBWSupported",       //  39

    "IsQoSConfigVLRateLimitSupported",                     //  40
    "IsQoSConfigVLAllocBWSupported",                       //  41
    nullptr,                                               //  42
    "IsCableInfoBankNumberSupported",                      //  43
    "IsDCQCNSupported",                                    //  44
    "IsAdaptiveTimeoutSupported",                          //  45
    "IsClassEnableSupported",                              //  46
    "IsExtendedAddressModeSupported",                      //  47

    "IsRouterLIDSupported",                                //  48
    "IsLinearForwardingTableSplitSupported",               //  49
    "IsDataTypetoSLMappingSupported",                      //  50
    "IsQoSConfigVLSupported",                              //  51
    "IsFastRecoverySupported",                             //  52
    "IsCreditWatchdogSupported",                           //  53
    "IsNVLFabricManagementSupported",                      //  54
    "IsNVLReductionManagementSupported",                   //  55

    "IsRailFilterSupported",                               //  56
    "IsNVLHBFConfigSupported",                             //  57
    "IsDataTypetoVLMappingSupported",                      //  58
    "IsContainAndDrainSupported",                          //  59
    "IsProfilesConfigSupported",                           //  60
    "IsBERConfigSupported",                                //  61
    "IsGlobalOOOV2Supported",                              //  62
    "IsVLBufferWeightSupported",                           //  63

    "IsEndPortPlaneFilterConfigSupported",                 //  64
    "IsEntryPlaneFilterSupported",                         //  65
    "IsBWUtiliztionSupported",                             //  66
    "IsExtendedSwitchInfoSupported",                       //  67
    "IsFabricManagerInfoSupported"                         //  68
    "IsExtPortInfoUnhealthyReasonSupported"                //  69
    "IsMulticastPrivateLinearForwardingSupported"          //  70
    "IsPortSLToMulticastPrivateLFTMapSupported"            //  71

    "IsChassisInfoSupported"                               //  72
    "IsPLRTriggerSupported"                                //  73
    "IsFwISSUSupported"                                    //  74
};

const std::vector<const char*> SimInfoData::GMPCapMaskBits = {
    "IsPortPowerStateSupported",                   //   0
    "IsDeviceSoftResetSupported",                  //   1
    "IsExtGPIOPortAccessSupported",                //   2
    "IsPhyConfigSupported",                        //   3
    "IsVendSpecLedControlSupported",               //   4
    "IsConfigSpaceAccessSupported",                //   5
    "IsPerVLCountersSupported",                    //   6
    "IsPerVLExtendedSupported",                    //   7

    "IsPortLLRStatisticsSupported",                //   8
    "IsCounterGroupsSupported",                    //   9
    "IsEnhancedConfigSpaceAccessSupported",        //  10
    "IsFlowMonitoringSupported",                   //  11
    "IsCongestionNotificationSupported",           //  12
    "IsVirtualSwitchSupported",                    //  13
    "IsPortXmitQLengthSupported",                  //  14
    "IsMulticastForwardingTableSupported",         //  15

    "IsPortXmitWaitVLExtendedSupported",           //  16
    "IsExtI2CPortAccessSupported",                 //  17
    "IsDiagnosticDataSupported",                   //  18
    "IsMaxRetransmissionRateSupported",            //  19
    "IsAccessRegisterGMPSupported",                //  20
    "IsPortHistogramsSupported",                   //  21
    "IsPoolHistogramsSupported",                   //  22
    "IsVPortCountersSupported",                    //  23

    "IsDiagnosticDataIdentificationSupported",     //  24
    "IsEnhPort0DeviceResetSupported",              //  25
    "IsDevInfoSupported",                          //  26
    "IsMADAddlStatusSupported",                    //  27
    "IsCableBurnSupported",                        //  28
    "IsMirroringEventsSupported",                  //  29
    "IsBERMonitoringSupported",                    //  30
    "IsBERMonitoringEventsSupported",              //  31

    "IsDCQCNSupported",                            //  32
    "IsPLRMaxRetransmissionRateSupported",         //  33
    "IsEffectiveCounterSupported",                 //  34
    "IsRawBerPerLaneSupported",                    //  35
    "IsAccessRegisterSupported",                   //  36
    "IsMirroringInfoSupported",                    //  37
    "IsExtendedAddressModeSupported",              //  38
    "IsSwitchNetworkInfoSupported",                //  39

    "CreditWatchdogTimeoutCounters",               //  40
    "FastRecoveryCountersSupported",               //  41
    "IsGeneralResetSupported",                     //  42
    "IsBufferHistogramSupported",                  //  43
    "IsPortsHistogramSupported",                   //  44
    "IsPacketBufferAllocControlSupported"          //  45
};

const std::vector<const char*> SimInfoData::PortInfoCapMaskBits = {
    nullptr,                                       //   0
    "IsSM",                                        //   1
    "IsNoticeSupported",                           //   2
    "IsTrapSupported",                             //   3
    "IsOptionalIPDSupported",                      //   4
    "IsAutomaticMigrationSupported",               //   5
    "IsSLMappingSupported",                        //   6
    "IsMKeyNVRAM",                                 //   7

    "IsPKeyNVRAM",                                 //   8
    "IsLEDInfoSupported",                          //   9
    "IsSMdisabled",                                //  10
    "IsSystemImageGUIDSupported",                  //  11
    "IsPKeySwitchExternalPortTrapSupported",       //  12
    "IsCableInfoSupported",                        //  13
    "IsExtendedSpeedsSupported",                   //  14
    "IsCapabilityMask2Supported",                  //  15

    "IsCommunicationManagementSupported",          //  16
    "IsSNMPTunnelingSupported",                    //  17
    "IsReinitSupported",                           //  18
    "IsDeviceManagementSupported",                 //  19
    "IsVendorClassSupported",                      //  20
    "IsDRNoticeSupported",                         //  21
    "IsCapabilityMaskNoticeSupported",             //  22
    "IsBootManagementSupported",                   //  23

    "IsLinkRoundTripLatencySupported",             //  24
    "IsClientReregistrationSupported",             //  25
    "IsOtherLocalChangesNoticeSupported",          //  26
    "IsLinkSpeedWidthPairsTableSupported",         //  27
    "IsVendorSpecificMadsTableSupported",          //  28
    "IsMulticastPKeyTrapSuppressionSupported",     //  29
    "IsMulticastFDBTopSupported",                  //  30
    "IsHierarchyInfoSupported"                     //  31
};

const std::vector<const char*> SimInfoData::PortInfoCapMask2Bits = {
    "IsSetNodeDescriptionSupported",               //   0
    "IsPortInfoExtendedSupported",                 //   1
    "IsVirtualizationSupported",                   //   2
    "IsSwitchPortStateTableSupported",             //   3
    "IsLinkWidth2xSupported",                      //   4
    "IsLinkSpeedHDRSupported",                     //   5
    "IsMKeyProtectBitsExtSupported",               //   6
    "IsEnhancedTrap128Supported",                  //   7

    "IsPartitionTopSupported",                     //   8
    "IsEnhancedQoSArbiterSupported",               //   9
    "IsLinkSpeedNDRSupported",                     //  10
    "IsExtendedSpeeds2Supported",                  //  11
    "IsLinkSpeedXDRSupported",                     //  12
    nullptr,                                       //  13
    nullptr,                                       //  14
    nullptr                                        //  15
};


int SimInfoDumpPy::Dump(const std::string & filename) {

    IBDIAG_ENTER

    std::ofstream  sout;
    int rc = m_ibdiag.OpenFile("Simulator Info (Python)", filename, sout, false, true);

    if (rc)
       IBDIAG_RETURN(rc);

    for (auto& x : m_ibdiag.GetDiscoverFabricPtr()->NodeByName) {
        IBNode* p_node = x.second;

        if (!p_node)
            continue;

        sout << std::setw(80) << "" << std::endl;

        sout << "Node Type: "   << nodetype2char(p_node->type)
             << (p_node->isSpecialNode() ? "(Special Node)" : "")
             << ", Node GUID: " << PTR(p_node->guid_get()) << std::setfill(' ')
             << ", Ports: "     << +p_node->numPorts
             << std::endl;

        sout << "Name: '"        << p_node->getName()
             << "', Alt: '"      << p_node->getAlternativeName()
             << std::endl;

        sout << std::setw(80) << "" << std::endl;

        this->DumpSMPCapMask(sout, *p_node);
        this->DumpGMPCapMask(sout, *p_node);

        this->DumpPortInfoAllCapMask(sout, *p_node);
    }

    m_ibdiag.CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SimInfoDumpPy::DumpPortInfoAllCapMask(std::ostream & sout, IBNode & node) {
    for (phys_port_t i = node.type == IB_SW_NODE ? 0 : 1; i <= node.numPorts; ++i) {
        const IBPort * p_port = node.getPort(i);

        if(!p_port)
            continue;

        if(i && !p_port->is_data_worthy())
            continue;

        SMP_PortInfo *p_info = m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);

        if (!p_info)
            continue;

        DumpPortInfoCapMask(sout, i, p_info, false);
        DumpPortInfoCapMask(sout, i, p_info, true);
    }
}

void SimInfoDumpPy::DumpPortInfoCapMask(std::ostream & sout, phys_port_t index, SMP_PortInfo *p_info, bool is_cap_mask2) {
    std::stringstream text;
    text << std::setw(12) << "" << "# ---------- Port[" << +index << "] CapMask"
         << (is_cap_mask2 ? "2" : "" ) << " ----------" << std::endl;
    text << std::setw(12) << "" << "caps = node.getPortInfo(index).get_capability_mask"
         << (is_cap_mask2 ? "2" : "" ) << "_bits()" << std::endl;

    const std::vector<const char *> &vec 
        = is_cap_mask2 ? SimInfoData::PortInfoCapMask2Bits : SimInfoData::PortInfoCapMaskBits;

    uint32_t capMask = is_cap_mask2 ? p_info->CapMsk2 : p_info->CapMsk;

    bool is_empty   = true;
    for(size_t i = 0; i < vec.size(); ++i ) {
        const char * name = vec[i];

        if (!name)
            continue;

        if (capMask & (1 << i)) {
            text << std::setw(12) << "" << "caps." << name
                 << std::setw(int(50lu - strlen(name)))
                 << " = 1" << std::endl;

            is_empty = false;
        }
    }

    if (!is_empty)
        sout << text.str();
}

void SimInfoDumpPy::DumpSMPCapMask(std::ostream & sout, IBNode & node) {
    sout << std::setw(8) << "" << "# ---------- SMP Caps ----------" << std::endl;
    for (size_t i = 0; i < SimInfoData::SMPCapMaskBits.size(); ++i) {
        const char * name = SimInfoData::SMPCapMaskBits[i];

        if (!name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedSMPCapability(&node, (u_int8_t)i)) {
            sout << std::setw(8) << "" << "node.getCapMaskSMPBits()." << name
                 << std::setw((int)(50lu - strlen(name))) << " = 1" << std::endl;
        }
    }
}

void SimInfoDumpPy::DumpGMPCapMask(std::ostream & sout, IBNode & node) {
    sout << std::setw(8) << "" << "# ---------- GMP Caps ----------" << std::endl;
    for (size_t i = 0; i < SimInfoData::GMPCapMaskBits.size(); ++i) {
        const char * name = SimInfoData::GMPCapMaskBits[i];

        if (!name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedGMPCapability(&node, (u_int8_t)i)) {
            sout << std::setw(8) << "" << "node.getCapMaskGMPBits()." << name
                 << std::setw((int)(50lu - strlen(name))) << " = 1" << std::endl;
        }
    }
}

std::string SimInfoDumpCPP::GetDumpedNodeType(const IBNode& node) const
{
    IBDIAG_ENTER

    string ret;

    if (node.type == IB_SW_NODE)
       ret = "SW";
    else if (node.type == IB_RTR_NODE)
       ret = "RTR";
    else if (!node.isSpecialNode())
       ret = "HCA";
    else if (node.getSpecialNodeType() == IB_SPECIAL_PORT_AN)
        ret = "AN";
    else
        ret = "Unknown";

    IBDIAG_RETURN(ret);
}

#define GENERATE_FUNCTION(func_name, body) sout  << std::endl << "void " << class_name <<  "::" \
    << this->GetDumpedNodeType(*p_node) << "::" << func_name \
    << "(IBMSNode *node) {" << std::endl; \
    this->body(sout, *p_node); \
    sout << std::endl << "}" << std::endl;

#define TO_HEX(x) "0x" << std::hex << +x << std::dec
#define START_OFFSET_LINE(blank) std::endl << std::setw(blank) << ""
#define START_LINE START_OFFSET_LINE(offset)
#define START_FUNC_BLOCK std::setw(offset) << ""

#define FUNC_BLOCK_DEF_MAD_BUFFER(buffer_type) sout << START_FUNC_BLOCK << #buffer_type << " mad_buffer = {0};";
#define DEF_MAD_BUFFER(buffer_type) sout << START_LINE << #buffer_type << " mad_buffer = {0};";
#define UNPACK_BUFFER(unpack_func, source) sout << START_LINE \
        << #unpack_func << "(&mad_buffer, reinterpret_cast<u_int8_t*>(&" << source << "));";
#define PACK_BUFFER(pack_func, source) sout << START_LINE \
        << #pack_func << "(&mad_buffer, reinterpret_cast<u_int8_t*>(&" << source << "));";
#define SET_MAD_VALUE(value) sout << START_LINE << "mad_buffer." \
        << #value << " = " << TO_HEX(p_source->value) << ";";
#define UNPACK_BUFFER_PTR(unpack_func, source) sout << START_LINE \
        << #unpack_func << "(&mad_buffer, reinterpret_cast<u_int8_t*>(" << source << "));";
#define PACK_BUFFER_PTR(pack_func, source) sout << START_LINE \
        << #pack_func << "(&mad_buffer, reinterpret_cast<u_int8_t*>(" << source << "));";


int SimInfoDumpCPP::Generate(const std::string &class_name, const std::set< u_int64_t >&simulator_guids)
{
    IBDIAG_ENTER

    string lower = class_name;
    transform(lower.begin(), lower.end(), lower.begin(), ::tolower);

    std::ofstream  sout;
    OutputControl::Identity identity(lower + "_generated.cpp", OutputControl::OutputControl_Flag_CustomFileName);
    int rc = m_ibdiag.OpenFile("Simulator generated code",
                               identity,
                               sout, false, true);
    if (rc)
       IBDIAG_RETURN(rc);

    this->PrintCopyright(sout);
    this->PrintIncludes(sout, lower);

    for (auto& x : m_ibdiag.GetDiscoverFabricPtr()->NodeByName) {
        IBNode* p_node = x.second;
        if (!p_node || !p_node->getInSubFabric())
            continue;

        // skip any node if it does not match nodes provided for simulator
        if (!simulator_guids.empty()) {
            if (simulator_guids.find(p_node->guid_get()) == simulator_guids.end())
                continue;
        }

        sout << std::endl << std::endl
             << "// Node Type: "   << nodetype2char(p_node->type)
             << (p_node->isSpecialNode() ? "(Special Node)" : "")
             << ", Node GUID: " << PTR(p_node->guid_get()) << std::setfill(' ')
             << ", Ports: "     << +p_node->numPorts
             << std::endl;

        sout << "// Name: '"        << p_node->getName()
             << "', Alt: '"      << p_node->getAlternativeName();

        //NodeInfo MAD
        GENERATE_FUNCTION("setNodeInfo", GenerateNodeInfo);
        //PortInfo MAD
        GENERATE_FUNCTION("setPortInfo", GeneratePortInfo);
        //VS GeneralInfo GMP MAD
        GENERATE_FUNCTION("setGMPCapability", GenerateGMPCapabilityMask);
        //VS GeneralInfoSMP MAD
        GENERATE_FUNCTION("setSMPCapability", GenerateSMPCapabilityMask);
        GENERATE_FUNCTION("setFWInfo", GenerateFWInfo);
        GENERATE_FUNCTION("setHWInfo", GenerateHWInfo);
        GENERATE_FUNCTION("setSWInfo", GenerateSWInfo);
        //PortInfoExtended MAD
        GENERATE_FUNCTION("setPortInfoExtended", GeneratePortInfoExtended);
        //ExtendedNodeInfo MAD
        GENERATE_FUNCTION("setExtendedNodeInfo", GenerateExtendedNodeInfo);
        //ExtendedPortInfo MAD
        GENERATE_FUNCTION("setExtendedPortInfo", GenerateExtendedPortInfo);
        //Adaptive Routing Info (ARInfo)MAD
        GENERATE_FUNCTION("setArInfo", GenerateARInfo);

        //ExtendedSwitchInfo
        GENERATE_FUNCTION("setExtendedSwitchInfo", GenerateExtendedSwitchInfo);
        //SwitchInfo MAD
        GENERATE_FUNCTION("setSwitchInfo", GenerateSwitchInfo);
    }

    m_ibdiag.CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SimInfoDumpCPP::GeneratePortInfoExtended(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    const SMP_PortInfoExtended *p_source = nullptr;
    for (phys_port_t i = 1; i <= node.numPorts; i++) {
        const IBPort *p_port = node.getPort(i);
        if (!p_port || !p_port->is_data_worthy())
            continue;

        p_source = m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPPortInfoExtended(p_port->createIndex);
        if (!p_source)
            continue;

        break;
    }

    this->GeneratePortInfoExtended(sout, p_source);

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GeneratePortInfoExtended(std::ostream &sout,
                                              const SMP_PortInfoExtended *p_source) const
{
    IBDIAG_ENTER;

    if (!p_source) {
        sout << "// Failed to dump PortInfoExtended MAD. Null pointer is provided.";
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    sout << START_FUNC_BLOCK << "for (size_t i = 0; i < node->nodePortsInfoExt.size(); i++) {";

    offset = 8;
    DEF_MAD_BUFFER(SMP_PortInfoExtended);
    UNPACK_BUFFER(SMP_PortInfoExtended_unpack, "node->nodePortsInfoExt[i]");
    sout << std::endl;

    //TODO consider special and FNM ports
    SET_MAD_VALUE(CapMsk);
    SET_MAD_VALUE(FECModeActive);
    SET_MAD_VALUE(FDRFECModeSupported);
    SET_MAD_VALUE(EDRFECModeSupported);
    SET_MAD_VALUE(HDRFECModeSupported);
    SET_MAD_VALUE(NDRFECModeSupported);

    sout << std::endl;
    PACK_BUFFER(SMP_PortInfoExtended_pack, "node->nodePortsInfoExt[i]");

    offset = 4;
    sout << START_LINE << "}";

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateExtendedPortInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    ext_port_info_map fnm_ports;
    const SMP_MlnxExtPortInfo *p_source = nullptr;
    for (phys_port_t i = 1; i <= node.numPorts; i++) {
        const IBPort *p_port = node.getPort(i);
        if (!p_port || !p_port->is_data_worthy())
            continue;

        SMP_MlnxExtPortInfo *ptr =
                m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPMlnxExtPortInfo(p_port->createIndex);
        if (!ptr)
            continue;

        if (!p_port->isFNMFamilyPort()) {
            if (!p_source)
                p_source = ptr;

            continue;
        }

        fnm_ports[p_port->num] = ptr;
    }

    this->GenerateExtendedPortInfo(sout, p_source, fnm_ports);

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateExtendedPortInfo(std::ostream &sout, const SMP_MlnxExtPortInfo *p_info,
                                              const ext_port_info_map &fnm_ports) const
{
   IBDIAG_ENTER;

    if (!p_info) {
        sout << "// Failed to dump ExtendedPortInfo MAD. Null pointer is provided.";
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    sout << START_FUNC_BLOCK << "for (size_t i = 0; i < node->nodeExtPortsInfo.size(); i++) {";

    offset = 8;
    DEF_MAD_BUFFER(SMP_MlnxExtPortInfo);
    UNPACK_BUFFER(SMP_MlnxExtPortInfo_unpack, "node->nodeExtPortsInfo[i]");
    sout << std::endl;

    sout << START_LINE << "switch (i) {";

    for (const auto &entry : fnm_ports) {
        if (!entry.second)
            continue;

        offset = 8;
        sout << START_LINE << "case " << +entry.first << ':';

        offset = 12;
        this->GenerateExtendedPortInfo(sout, offset, entry.second);

        sout << START_LINE << "break;";
    }

    offset = 8;
    sout << START_LINE << "default:";

    offset = 12;
    this->GenerateExtendedPortInfo(sout, offset, p_info);

    offset = 8;
    sout << START_LINE << '}';

    sout << std::endl;
    PACK_BUFFER(SMP_MlnxExtPortInfo_pack, "node->nodeExtPortsInfo[i]");

    offset = 4;
    sout << START_LINE << "}";

   IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateExtendedPortInfo(std::ostream &sout, int offset,
                                              const SMP_MlnxExtPortInfo *p_source) const
{
    IBDIAG_ENTER;

    SET_MAD_VALUE(LinkSpeedSupported);
    SET_MAD_VALUE(CapabilityMask);
    SET_MAD_VALUE(FDR10FECModeSupported);
    SET_MAD_VALUE(FDRFECModeSupported);
    SET_MAD_VALUE(EDR20FECModeSupported);
    SET_MAD_VALUE(EDRFECModeSupported);
    SET_MAD_VALUE(FDR10RetranSupported);
    SET_MAD_VALUE(FDRRetranSupported);
    SET_MAD_VALUE(EDR20RetranSupported);
    SET_MAD_VALUE(EDRRetranSupported);

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::PrintCopyright(std::ostream &sout) const
{
    IBDIAG_ENTER;

    std::time_t t = std::time(nullptr);
    std::tm *const p_timeInfo = std::localtime(&t);

    sout << "/*" << std::endl
         << "* SPDX-FileCopyrightText: Copyright (c) "
         <<  (p_timeInfo ? p_timeInfo->tm_year + 1900 : 2024)
         << " NVIDIA CORPORATION & AFFILIATES. All rights reserved" << std::endl
         << "* SPDX-License-Identifier: LicenseRef-NvidiaProprietary" << std::endl
         << "*" << std::endl
         << "* NVIDIA CORPORATION, its affiliates and licensors retain all intellectual" << std::endl
         << "* property and proprietary rights in and to this material, related" << std::endl
         << "* documentation and any modifications thereto. Any use, reproduction," << std::endl
         << "* disclosure or distribution of this material and related documentation" << std::endl
         << "* without an express license agreement from NVIDIA CORPORATION or" << std::endl
         << "* its affiliates is strictly prohibited." << std::endl
         << "*/" << std::endl << std::endl;

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::PrintIncludes(std::ostream &sout, const std::string &name) const
{
    IBDIAG_ENTER;

    sout << "#include \"" << name << ".h\"" << std::endl;
    sout << "#include \"../../core/node.h\"" << std::endl;
    sout << "#include <infiniband/ibdiag/ibdiag_fabric.h>" << std::endl;

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateSwitchInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    if (node.type != IB_SW_NODE) {
        sout << "// The node=" << node.getName() << " is not a switch";
        IBDIAG_RETURN_VOID;
    }

    const SMP_SwitchInfo *p_source =
                this->m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPSwitchInfo(node.createIndex);
    if (!p_source) {
        sout << "// Failed to get SwitchInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    FUNC_BLOCK_DEF_MAD_BUFFER(SMP_SwitchInfo);
    UNPACK_BUFFER(SMP_SwitchInfo_unpack, "node->switchInfo");
    sout << std::endl;

    SET_MAD_VALUE(LinearFDBCap);
    SET_MAD_VALUE(RandomFDBCap);
    SET_MAD_VALUE(MCastFDBCap);
    SET_MAD_VALUE(OptimizedSLVLMapping);
    SET_MAD_VALUE(LidsPerPort);
    SET_MAD_VALUE(PartEnfCap);
    SET_MAD_VALUE(InbEnfCap);
    SET_MAD_VALUE(OutbEnfCap);
    SET_MAD_VALUE(FilterRawInbCap);
    SET_MAD_VALUE(FilterRawOutbCap);
    SET_MAD_VALUE(ENP0);

    sout << std::endl;
    PACK_BUFFER(SMP_SwitchInfo_pack, "node->switchInfo");
    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateExtendedNodeInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;


    if (!this->m_ibdiag.GetCapabilityModulePtr()->IsSupportedSMPCapability(
                &node, EnSMPCapIsExtendedNodeInfoSupported)) {
        sout << "// The ExtendedNodeInfo mad is not supported on the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    const ib_extended_node_info* p_source =
                this->m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPExtNodeInfo(node.createIndex);
    if (!p_source) {
        sout << "// Failed to get VS ExtendedNodeInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    FUNC_BLOCK_DEF_MAD_BUFFER(ib_extended_node_info);
    UNPACK_BUFFER(ib_extended_node_info_unpack, "node->extNodeInfo");
    sout << std::endl;

    SET_MAD_VALUE(sl2vl_cap);
    SET_MAD_VALUE(node_type_extended);
    SET_MAD_VALUE(num_pcie);
    SET_MAD_VALUE(num_oob);

    sout << std::endl;
    PACK_BUFFER(ib_extended_node_info_pack, "node->extNodeInfo");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateARInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    if (node.type != IB_SW_NODE) {
        sout << "// The node=" << node.getName() << " is not a switch";
        IBDIAG_RETURN_VOID;
    }

    adaptive_routing_info* p_source =
                this->m_ibdiag.GetIBDMExtendedInfoPtr()->getARInfo(node.createIndex);
    if (!p_source) {
        sout << "// Failed to get VS ArInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    FUNC_BLOCK_DEF_MAD_BUFFER(adaptive_routing_info);
    UNPACK_BUFFER(adaptive_routing_info_unpack, "node->AdaptiveRoutingInfo");
    sout << std::endl;

    SET_MAD_VALUE(dyn_cap_calc_sup);
    SET_MAD_VALUE(by_transp_cap);
    SET_MAD_VALUE(by_sl_cap);
    SET_MAD_VALUE(glb_groups);
    SET_MAD_VALUE(is4_mode);
    SET_MAD_VALUE(direction_num_sup);
    SET_MAD_VALUE(group_table_copy_sup);
    SET_MAD_VALUE(is_ar_trials_supported);
    SET_MAD_VALUE(is_fr_sup);
    SET_MAD_VALUE(is_frn_sup);
    SET_MAD_VALUE(is_arn_sup);
    SET_MAD_VALUE(string_width_cap);
    SET_MAD_VALUE(group_table_cap);
    SET_MAD_VALUE(sub_grps_supported);
    SET_MAD_VALUE(rn_version_cap);
    SET_MAD_VALUE(ar_version_cap);
    SET_MAD_VALUE(is_hbf_supported);
    SET_MAD_VALUE(is_whbf_supported);
    SET_MAD_VALUE(is_symmetric_hash_supported);
    SET_MAD_VALUE(is_dceth_hash_supported);
    SET_MAD_VALUE(is_bth_dqp_hash_supported);
    SET_MAD_VALUE(is_pfrn_supported);

    sout << std::endl;
    PACK_BUFFER(adaptive_routing_info_pack, "node->AdaptiveRoutingInfo");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateExtendedSwitchInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    if (node.type != IB_SW_NODE) {
        sout << "// The node=" << node.getName() << " is not a switch";
        IBDIAG_RETURN_VOID;
    }

    const SMP_ExtendedSwitchInfo *p_source =
                    this->m_ibdiag.GetIBDMExtendedInfoPtr()->getSMP_ExtendedSwitchInfo(node.createIndex);
    if (!p_source) {
        sout << "// Failed to get VS ExtendedSwitchInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    FUNC_BLOCK_DEF_MAD_BUFFER(SMP_ExtendedSwitchInfo);
    UNPACK_BUFFER_PTR(SMP_ExtendedSwitchInfo_unpack, "node->getExtendedSwitchInfo()");
    sout << std::endl;

    SET_MAD_VALUE(turbo_path_cap);

    sout << std::endl;
    PACK_BUFFER_PTR(SMP_ExtendedSwitchInfo_pack, "node->getExtendedSwitchInfo()");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateSWInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    if (node.isSpecialNode()) {
        sout << "// The node=" << node.getName()
             << " is special and it should not support VS GeneralInfo mad.";
        IBDIAG_RETURN_VOID;
    }

    const VendorSpec_GeneralInfo *p_general_info =
        this->m_ibdiag.GetIBDMExtendedInfoPtr()->getVSGeneralInfo(node.createIndex);
    if (!p_general_info) {
        sout << "// Failed to get VS GeneralInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    const SWInfo_Block_Element *p_source = &p_general_info->SWInfo;

    FUNC_BLOCK_DEF_MAD_BUFFER(SWInfo_Block_Element);
    UNPACK_BUFFER(SWInfo_Block_Element_unpack, "node->swInfo");
    sout << std::endl;

    SET_MAD_VALUE(Major);
    SET_MAD_VALUE(Minor);
    SET_MAD_VALUE(SubMinor);

    sout << std::endl;
    PACK_BUFFER(SWInfo_Block_Element_pack, "node->swInfo");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateFWInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

     if (node.isSpecialNode()) {
        sout << "// The node=" << node.getName()
             << " is special and it should not support VS General Info mad.";
        IBDIAG_RETURN_VOID;
     }

    const VendorSpec_GeneralInfo *p_general_info =
        this->m_ibdiag.GetIBDMExtendedInfoPtr()->getVSGeneralInfo(node.createIndex);
    if (!p_general_info) {
        sout << "// Failed to get VS GeneralInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    const FWInfo_Block_Element *p_source = &p_general_info->FWInfo;

    FUNC_BLOCK_DEF_MAD_BUFFER(FWInfo_Block_Element)
    UNPACK_BUFFER(FWInfo_Block_Element_unpack, "node->fwInfo");
    sout << std::endl;

    SET_MAD_VALUE(Major);
    SET_MAD_VALUE(Minor);
    SET_MAD_VALUE(SubMinor);

    SET_MAD_VALUE(Extended_Major);
    SET_MAD_VALUE(Extended_Minor);
    SET_MAD_VALUE(Extended_SubMinor);

    sout << std::endl;
    PACK_BUFFER(FWInfo_Block_Element_pack, "node->fwInfo");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateHWInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    if (node.isSpecialNode()) {
        sout << "// The node=" << node.getName()
             << " is special and it should not support VS General Info mad.";
        IBDIAG_RETURN_VOID;
     }

    const VendorSpec_GeneralInfo *p_general_info =
        this->m_ibdiag.GetIBDMExtendedInfoPtr()->getVSGeneralInfo(node.createIndex);
    if (!p_general_info) {
        sout << "// Failed to get VS GeneralInfo mad for the node=" << node.getName();
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    const HWInfo_Block_Element *p_source = &p_general_info->HWInfo;

    FUNC_BLOCK_DEF_MAD_BUFFER(HWInfo_Block_Element)
    UNPACK_BUFFER(HWInfo_Block_Element_unpack, "node->hwInfo");
    sout << std::endl;

    SET_MAD_VALUE(technology);
    SET_MAD_VALUE(DeviceHWRevision);
    //SET_MAD_VALUE(manufacturing_base_mac_32_47);
    //SET_MAD_VALUE(manufacturing_base_mac_0_31);
    SET_MAD_VALUE(hw_dev_id);
    //SET_MAD_VALUE(pvc);
    //SET_MAD_VALUE(num_ports);

    sout << std::endl;
    PACK_BUFFER(HWInfo_Block_Element_pack, "node->hwInfo");

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateSMPCapabilityMask(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

#define offset 4
    capability_mask_t smp_cap;
    int rc = this->m_ibdiag.GetCapabilityModulePtr()->GetCapability(const_cast<IBNode*>(&node),
                                                          false, smp_cap);
    if (rc) {
        sout << "// Failed to get SMP capabilites for the node=" << node.getName()
             << " with rc=" << rc;
        IBDIAG_RETURN_VOID;
    }

    sout << START_FUNC_BLOCK << "// ---------- SMP Caps ----------";
    for (size_t i = 0; i < SimInfoData::SMPCapMaskBits.size(); ++i) {
        const char *p_name = SimInfoData::SMPCapMaskBits[i];
        if (!p_name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedSMPCapability(&node, (u_int8_t)i)) {
            sout << START_LINE << "// " << p_name;
        }
    }

    sout << std::endl;
    DEF_MAD_BUFFER(GeneralInfoCapabilityMask);
    UNPACK_BUFFER(GeneralInfoCapabilityMask_unpack, "node->capMaskSMP");
    sout << std::endl;

    for (size_t i = 0; i < ARRAY_SIZE(smp_cap.mask); i++)
        sout << START_LINE
             << "mad_buffer.capability" << i << " = " << TO_HEX(smp_cap.mask[i])
             << ";";

    sout << std::endl;
    PACK_BUFFER(GeneralInfoCapabilityMask_pack, "node->capMaskSMP");
#undef offset

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateGMPCapabilityMask(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

#define offset 4
    if (node.isSpecialNode()) {
        sout << "// The node=" << node.getName()
             << " is special and it should not support GMP capabilities.";
        IBDIAG_RETURN_VOID;
    }

    capability_mask_t gmp_cap;
    int rc = this->m_ibdiag.GetCapabilityModulePtr()->GetCapability(const_cast<IBNode*>(&node),
                                                          true, gmp_cap);
    if (rc) {
        sout << "// Failed to get GMP capabilites for the node=" << node.getName()
             << " with rc=" << rc;
        IBDIAG_RETURN_VOID;
    }

    sout << START_FUNC_BLOCK << "// ---------- GMP Caps ----------";
    for (size_t i = 0; i < SimInfoData::GMPCapMaskBits.size(); ++i) {
        const char *p_name = SimInfoData::GMPCapMaskBits[i];
        if (!p_name)
            continue;

        if (m_ibdiag.GetCapabilityModulePtr()->IsSupportedGMPCapability(&node, (u_int8_t)i)) {
            sout << START_LINE << "// " << p_name;
        }
    }

    sout << std::endl;
    DEF_MAD_BUFFER(GeneralInfoCapabilityMask);
    UNPACK_BUFFER(GeneralInfoCapabilityMask_unpack, "node->capMaskGMP");
    sout << std::endl;

    for (size_t i = 0; i < ARRAY_SIZE(gmp_cap.mask); i++)
        sout << START_LINE
             << "mad_buffer.capability" << i << " = " << TO_HEX(gmp_cap.mask[i])
             << ";";

    sout << std::endl;
    PACK_BUFFER(GeneralInfoCapabilityMask_pack, "node->capMaskGMP");
#undef offset

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GenerateNodeInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER;

    const SMP_NodeInfo *p_source =
                this->m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPNodeInfo(node.createIndex);
    if (!p_source) {
        sout << "#error \"Failed to get NodeInfo mad for the node=" << node.getName() << "\"";
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    FUNC_BLOCK_DEF_MAD_BUFFER(SMP_NodeInfo);
    UNPACK_BUFFER(SMP_NodeInfo_unpack, "node->nodeInfo");
    sout << std::endl;

    SET_MAD_VALUE(BaseVersion);
    SET_MAD_VALUE(ClassVersion);
    SET_MAD_VALUE(PartitionCap);
    SET_MAD_VALUE(DeviceID);
    SET_MAD_VALUE(revision);
    SET_MAD_VALUE(VendorID);

    sout << std::endl;
    PACK_BUFFER(SMP_NodeInfo_pack, "node->nodeInfo");

    sout << std::endl;
    sout << START_LINE << "node->resizePortPkeyTable(mad_buffer.PartitionCap);";

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GeneratePortInfo(std::ostream &sout, const IBNode &node) const
{
    IBDIAG_ENTER

    const SMP_PortInfo *p_info = nullptr;
    const SMP_PortInfo *p_zero = nullptr;
    port_info_map fnm_ports;
    phys_port_t from_index = (node.type == IB_SW_NODE ? 0 : 1);
    for (phys_port_t i = from_index; i <= node.numPorts; i++) {
        const IBPort *p_port = node.getPort(i);
        if (!p_port)
            continue;

        if (i && !p_port->is_data_worthy())
            continue;

        const SMP_PortInfo *ptr = m_ibdiag.GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);
        if (!ptr)
            continue;

        if (!i) {
            p_zero = ptr;
            continue;
        }

        if (i && !p_port->isFNMFamilyPort()) {
            if (!p_info)
                p_info = ptr;

            continue;
        }

        fnm_ports[p_port->num] = ptr;
    }

    if (node.type == IB_SW_NODE)
        this->GeneratePortInfoSW(sout, p_zero, p_info, fnm_ports);
    else
        this->GeneratePortInfoCA(sout, p_info);

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::PrintCapabilityMask(int offset, std::ostream &sout,
                                         u_int32_t cap_mask, bool is_cap_mask2) const
{
    IBDIAG_ENTER;

    const std::vector<const char *> &vec
        = is_cap_mask2 ? SimInfoData::PortInfoCapMask2Bits : SimInfoData::PortInfoCapMaskBits;

    sout << START_LINE << "// ---------- CapMask" << (is_cap_mask2 ? "2" : "");

    for (size_t i = 0; i < vec.size(); ++i) {
        const char *p_name = vec[i];
        if (!p_name)
            continue;

        if (cap_mask & (1 << i)) {
            sout << START_LINE << "// " << p_name
                 << std::setw(int(50lu - strlen(p_name)))
                 << " = 1";
        }
    }

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GeneratePortInfoSW(std::ostream &sout, const SMP_PortInfo *p_source,
                          int offset, bool print_cap) const
{
    IBDIAG_ENTER;

    if (print_cap)
        this->PrintCapabilityMask(offset, sout, p_source->CapMsk, false);

    SET_MAD_VALUE(CapMsk);

    if (print_cap) {
        sout << std::endl;
        this->PrintCapabilityMask(offset, sout, p_source->CapMsk2, true);
    }

    SET_MAD_VALUE(CapMsk2);
    sout << std::endl;

    //setting Link WIDTH
    SET_MAD_VALUE(LinkWidthSup);
    sout << std::endl;

    //setting Link Speed
    SET_MAD_VALUE(LinkSpeedExtSup2)
    SET_MAD_VALUE(LinkSpeedExtSup);
    SET_MAD_VALUE(LinkSpeedSup);

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GeneratePortInfoSW(std::ostream &sout, const SMP_PortInfo *p_zero,
                                        const SMP_PortInfo *p_info,
                                        const port_info_map &fnm_ports) const
{
    IBDIAG_ENTER;

    if (!p_zero || !p_info) {
        sout << "#error \"Failed to dump PortInfo MAD. Zero and a regular ports must be provided.\"";
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    sout << START_FUNC_BLOCK << "for (size_t i = 0; i < node->nodePortsInfo.size(); i++) {";
    offset = 8;
    DEF_MAD_BUFFER(SMP_PortInfo);
    UNPACK_BUFFER(SMP_PortInfo_unpack, "node->nodePortsInfo[i]");
    sout << std::endl;

    sout << START_LINE << "if (!i) {";
    offset =  12;

    this->GeneratePortInfoSW(sout, p_zero, offset, true);

    for (const auto& entry : fnm_ports) {
        if (!entry.second)
            continue;

        offset = 8;
        sout << START_LINE << "} else if (i == " << +entry.first << ") {";
        offset  = 12;

        this->GeneratePortInfoSW(sout, entry.second, offset);
    }

    offset = 8;
    sout << START_LINE << "} else {";
    offset  = 12;

    this->GeneratePortInfoSW(sout, p_info, offset);

    offset = 8;
    sout << START_LINE << "}";

    sout << std::endl;
    PACK_BUFFER(SMP_PortInfo_pack, "node->nodePortsInfo[i]");

    offset = 4;
    sout << START_LINE << "}";

    IBDIAG_RETURN_VOID;
}

void SimInfoDumpCPP::GeneratePortInfoCA(std::ostream &sout, const SMP_PortInfo *p_source) const
{
    IBDIAG_ENTER;

    if (!p_source) {
        sout << "#error \"Failed to dump PortInfo MAD. Null pointer is provided.\"";
        IBDIAG_RETURN_VOID;
    }

    int offset = 4;
    sout << START_FUNC_BLOCK << "for (size_t i = 0; i < node->nodePortsInfo.size(); i++) {";

    offset = 8;
    DEF_MAD_BUFFER(SMP_PortInfo);
    UNPACK_BUFFER(SMP_PortInfo_unpack, "node->nodePortsInfo[i]");
    sout << std::endl;

    this->PrintCapabilityMask(offset, sout, p_source->CapMsk, false);
    SET_MAD_VALUE(CapMsk);
    sout << std::endl;

    this->PrintCapabilityMask(offset, sout, p_source->CapMsk2, true);
    SET_MAD_VALUE(CapMsk2);
    sout << std::endl;

    //setting Link WIDTH
    SET_MAD_VALUE(LinkWidthSup);
    sout << std::endl;

    //setting Link Speed
    SET_MAD_VALUE(LinkSpeedExtSup2)
    SET_MAD_VALUE(LinkSpeedExtSup);
    SET_MAD_VALUE(LinkSpeedSup);
    sout << std::endl;

    PACK_BUFFER(SMP_PortInfo_pack, "node->nodePortsInfo[i]");

    offset = 4;
    sout << START_LINE << "}";

    IBDIAG_RETURN_VOID;
}

