/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <functional>
#include <iomanip>
#include  <sys/socket.h>
#include  <arpa/inet.h>
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibutils/output_stream_primitives.h"

extern IBDiagClbck ibDiagClbck;

//Router related definitions
#define IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK 8
#define IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK 4
#define NUM_OF_ROUTER_TABLE_ENTRIES(tbl_top, num_elements_in_block) \
    ((tbl_top + num_elements_in_block - 1) / num_elements_in_block)


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpNodesInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    std::ios_base::fmtflags oldflags = sout.flags();

    struct VendorSpec_GeneralInfo *p_curr_general_info = NULL;
    char buffer[2096];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        capability_mask_t capability_mask_gmp;
        int rc = this->capability_module.GetCapability(p_curr_node, true, capability_mask_gmp);
        p_curr_general_info = this->fabric_extended_info.getVSGeneralInfo(i);
        if (!p_curr_general_info && rc != IBDIAG_SUCCESS_CODE) //info not exist
            continue;

        sout << "-------------------------------------------------------" << endl;
        sout << "Node Name=" << p_curr_node->getName() << endl;
        sout << "-------------------------------------------------------" << endl;
        sout << "GUID=0x" << std::hex << std::setfill('0') << std::setw(16)
             << p_curr_node->guid_get() << endl;

        if (p_curr_general_info) {
            string psid = (char *)p_curr_general_info->FWInfo.PSID.PSID;
            snprintf(buffer, sizeof(buffer),
                     "HWInfo_DeviceID=" U16H_FMT "\n"    \
                     "HWInfo_DeviceHWRevision=" U16H_FMT "\n"    \
                     "HWInfo_technology=" U32D_FMT "\n"    \
                     "HWInfo_UpTime=" U32H_FMT "\n"  \
                     "FWInfo_Version=%u.%u.%u\n"    \
                     "FWInfo_secure_signed_debug_dev_fw=%u %u %u %u\n" \
                     "FWInfo_BuildID=" U32H_FMT "\n" \
                     "FWInfo_Date=%02x.%02x.%04x\n"    \
                     "FWInfo_Hour=%02x:%02x\n"    \
                     "FWInfo_PSID=" STR_FMT  "\n"    \
                     "FWInfo_INI_File_Version=" U32H_FMT "\n"    \
                     "FWInfo_Extended_Version=%u.%u.%u\n"  \
                     "FWInfo_isfu_major=" U16H_FMT "\n" \
                     "FWInfo_sec_boot=%u\n"\
                     "FWInfo_life_cycle=%u\n"\
                     "SWInfo_Version=%u.%u.%u\n",
                     p_curr_general_info->HWInfo.DeviceID,
                     p_curr_general_info->HWInfo.DeviceHWRevision,
                     p_curr_general_info->HWInfo.technology,
                     p_curr_general_info->HWInfo.UpTime,
                     p_curr_general_info->FWInfo.Major,
                     p_curr_general_info->FWInfo.Minor,
                     p_curr_general_info->FWInfo.SubMinor,
                     p_curr_general_info->FWInfo.secure_fw,
                     p_curr_general_info->FWInfo.signed_fw,
                     p_curr_general_info->FWInfo.debug_fw,
                     p_curr_general_info->FWInfo.dev_fw,
                     p_curr_general_info->FWInfo.BuildID,
                     p_curr_general_info->FWInfo.Month,
                     p_curr_general_info->FWInfo.Day,
                     p_curr_general_info->FWInfo.Year,
                     ((u_int8_t *)&p_curr_general_info->FWInfo.Hour)[1],
                     ((u_int8_t *)&p_curr_general_info->FWInfo.Hour)[0],
                     (psid == "") ? "UNKNOWN" : psid.c_str(),
                     p_curr_general_info->FWInfo.INI_File_Version,
                     p_curr_general_info->FWInfo.Extended_Major,
                     p_curr_general_info->FWInfo.Extended_Minor,
                     p_curr_general_info->FWInfo.Extended_SubMinor,
                     p_curr_general_info->FWInfo.isfu_major,
                     p_curr_general_info->FWInfo.sec_boot,
                     p_curr_general_info->FWInfo.life_cycle,
                     p_curr_general_info->SWInfo.Major,
                     p_curr_general_info->SWInfo.Minor,
                     p_curr_general_info->SWInfo.SubMinor);
            sout << buffer;
        } else {
            sout << "HWInfo_DeviceID=N/A" << endl
                 << "HWInfo_DeviceHWRevision=N/A" << endl
                 << "HWInfo_technology=N/A" << endl
                 << "HWInfo_UpTime=N/A" << endl
                 << "FWInfo_SubMinor=N/A" << endl
                 << "FWInfo_Minor=N/A" << endl
                 << "FWInfo_Major=N/A" << endl
                 << "FWInfo_secure_signed_debug_dev_fw=N/A" << endl
                 << "FWInfo_BuildID=N/A" << endl
                 << "FWInfo_Year=N/A" << endl
                 << "FWInfo_Day=N/A" << endl
                 << "FWInfo_Month=N/A" << endl
                 << "FWInfo_Hour=N/A" << endl
                 << "FWInfo_PSID=N/A" << endl
                 << "FWInfo_INI_File_Version=N/A" << endl
                 << "FWInfo_Extended_Major=N/A" << endl
                 << "FWInfo_Extended_Minor=N/A" << endl
                 << "FWInfo_Extended_SubMinor=N/A" << endl
                 << "FWInfo_isfu_major=N/A" << endl
                 << "FWInfo_sec_boot=N/A" << endl
                 << "FWInfo_life_cycle=N/A" << endl
                 << "SWInfo_SubMinor=N/A" << endl
                 << "SWInfo_Minor=N/A" << endl
                 << "SWInfo_Major=N/A" << endl;
        }

        for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS;
             capability_field++) {
                 sout << "CapabilityMask_" << std::dec << capability_field << "=";
                 if (rc == IBDIAG_SUCCESS_CODE)
                     sout << "0x" << std::hex << std::setfill('0') << std::setw(8)
                          << capability_mask_gmp.mask[capability_field] << endl;
                 else
                     sout << "N/A" << endl;
        }

        sout << endl;
    }

    sout.flags(oldflags);

    IBDIAG_RETURN_VOID;
}


//*****************************************************************************
//   Functions for printing to mlnx_cntrs file
//*****************************************************************************
void IBDiag::DumpDiagnosticCountersDescriptionP0(ofstream &sout)
{
    IBDIAG_ENTER;
    sout << "#---------- Transport errors and flows (Page0 Ver" << PAGE0_LATEST_VER <<
            ")----------------------------" << endl;
    sout << "#rq_num_lle: Responder - number of local length errors" << endl;
    sout << "#sq_num_lle: Requester - number of local length errors" << endl;
    sout << "#rq_num_lqpoe: Responder - number local QP operation error" <<
        endl;
    sout << "#sq_num_lqpoe: Requester - number local QP operation error" <<
        endl;
    sout << "#rq_num_leeoe: Responder - number local EE operation error" <<
        endl;
    sout << "#sq_num_leeoe: Requester - number local EE operation error" <<
        endl;
    sout << "#rq_num_lpe: Responder - number of local protection errors" <<
        endl;
    sout << "#sq_num_lpe: Requester - number of local protection errors" <<
        endl;
    sout << "#rq_num_wrfe: Responder - number of WR flushed errors" << endl;
    sout << "#sq_num_wrfe: Requester - number of WR flushed errors" << endl;
    sout << "#sq_num_mwbe: Requester - number of memory window bind errors" <<
        endl;
    sout << "#sq_num_bre: Requester - number of bad response errors" << endl;
    sout << "#rq_num_lae: Responder - number of local access errors" << endl;
    sout << "#rq_num_rire: Responder - number of remote invalid request"  \
            " errors" << endl;
    sout << "#sq_num_rire: Requester - number of remote invalid request"  \
            " errors" << endl;
    sout << "#rq_num_rae: Responder - number of remote access errors" << endl;
    sout << "#sq_num_rae: Requester - number of remote access errors" << endl;
    sout << "#rq_num_roe: Responder - number of remote operation errors" <<
        endl;
    sout << "#sq_num_roe: Requester - number of remote operation errors" <<
        endl;
    sout << "#sq_num_rnr: Responder - number of RNR Naks received" << endl;
    sout << "#rq_num_oos: Responder - number of out of sequence requests"  \
            " received" << endl;
    sout << "#sq_num_oos: Requester - number of out of sequence Naks "  \
            "received" << endl;
    sout << "#rq_num_dup: Responder - number of duplicate requests received" <<
        endl;
    sout << "#sq_num_to: Requester - number of time out received" << endl;
    sout << "#sq_num_tree: Requester - number of transport retries exceeded"  \
            " errors" << endl;
    sout << "#sq_num_rree: Requester - number of RNR nak retries exceeded"  \
            " errors" << endl;
    sout << "#sq_num_rabrte: Requester - number of remote aborted errors" <<
        endl;
    sout << "#rq_num_mce: Responder - number of bad multicast packets"  \
            " received"  << endl;
    sout << "#rq_num_retrans_rsync: Responder - number of retransmission"  \
            " RESYNC operations" << endl;
    sout << "#sq_num_retrans_rsync: Requester - number of retransmission"  \
            "  RESYNC operations" << endl;
    sout << "#sq_num_ldb_drops: Requester - number of non-fatal ack/response"  \
            " mismatches" << endl;

}


void IBDiag::DumpDiagnosticCountersDescriptionP1(ofstream &sout)
{
    IBDIAG_ENTER;

    sout << "#---------- HCA Extended Flows (Page1 Ver" << PAGE1_LATEST_VER <<  \
        ")----------------------------" << endl;

    sout << "#rq_num_sig_err: Responder - number of wire signature errors" << endl;
    sout << "#sq_num_sig_err: Requester - number of memory signature errors" << endl;
    sout << "#sq_num_cnak: Requester - number of DC Connect Nacks received" << endl;
    sout << "#sq_reconnect: Requester - number of DC reconnect requests sent" << endl;
    sout << "#sq_reconnect_ack: Requester - number of DC reconnect acks received" << endl;
    sout << "#rq_open_gb: Responder - number of DC ghost busters opened" << endl;
    sout << "#rq_num_no_dcrs: Responder-number of loopback Connects which"  \
            " got no DCRs" << endl;
    sout << "#rq_num_cnak_sent: Responder-number of loopback Connects"  \
            " which got no DCRs and sent a CNAK" << endl;
    sout << "#sq_reconnect_ack_bad: Requester-number of DC reconnect acks"  \
            " dropped" << endl;
    sout << "#rq_open_gb_cnak: Responder-number of DC ghost busters open"  \
            " required/connect while connect with ghost buster - but no"  \
            " DCRs available" << endl;
    sout << "#rq_gb_trap_cnak: Responder-number of cnaks sent on connect"  \
            " packet trapped at ghost buster" << endl;
    sout << "#rq_not_gb_connect: Responder-connect while connected, no"  \
            " ghost buster involved (dropped disconnect)" << endl;
    sout << "#rq_not_gb_reconnect: Responder-reconnect without ghost buster"  \
            " received = ghost buster installer " << endl;
    sout << "#rq_curr_gb_connect: Responder-connect received trapped by"  \
            " ghost buster" << endl;
    sout << "#rq_curr_gb_reconnect: Responder-ghost buster reactivated"  \
            " (ghost buster installer while ghost buster exists)" << endl;
    sout << "#rq_close_non_gb_gc: Responder-number of regular DCRs closed"  \
            " due to garbage collection" << endl;
    sout << "#rq_dcr_inhale_events: Responder-number of DCR inhale"  \
            " events triggered" << endl;
    sout << "#rq_state_active_gb: Responder-number of currently open"  \
            " ghost busters" << endl;
    sout << "#rq_state_avail_dcrs: Responder-number of currently"  \
            " available DCRs" << endl;
    sout << "#rq_state_dcr_lifo_size: Responder-current DCR"  \
            " stack size" << endl;
    sout << "#sq_cnak_drop: Requester - number of CNAKs dropped" << endl;
    sout << "#minimum_dcrs: Responder - lowest level of DCRs observed" << endl;
    sout << "#maximum_dcrs: Responder - maximum number of used DCRs observed" << endl;
    sout << "#max_cnak_fifo_size: Responder - highest number of pending CNAKs" << endl;
    sout << "#rq_num_dc_cacks: Responder - number of connect-ack packets sent" << endl;
    sout << "#sq_num_dc_cacks: Requester - number of connect-ack packets received" << endl;

    sout << "#rx_rsp_ooo_packet: Responder - number of received OOO packets (RDMA " \
            "write only, within PSN window)" << endl;
    sout << "#rx_req_ooo_packet: Requester - number of received OOO packets" << endl;
    sout << "#tx_reached_max_outstanding_psn: Number of events of transmitter reaching max " \
            "outstanding PSN and therefore pausing transmission. Includes OOO and non-OOO." << endl;
    sout << "#tx_pause_on_strong_ordering: Number of events of transmitter pausing transmission" \
            " due to fence between write and non-write WQEs." << endl;

    // ODP counters
    sout << "#odp_fault_discard_eq_overrun: Number of PAGE FAULTs detected, but no event sent due to EQ full" << endl;
    sout << "#odp_fault_discard_other: Number of PAGE FAULTs detected, but no event sent due to no internal resources" << endl;
    sout << "#odp_event_posted: Number of PAGE FAULT events posted" << endl;
    sout << "#odp_event_last_posted: Number of PAGE FAULT events posted with last indication" << endl;
    sout << "#odp_resume_received: Number of PAGE FAULT RESUME commands received" << endl;
    sout << "#odp_resume_error_received: Number of PAGE FAULT RESUME commands received with error indication" << endl;
    sout << "#odp_tx_req: Number of ODP faults triggered by TX requestor gather (e.g., outbound SEND / WRITE operations)" << endl;
    sout << "#odp_rx_res: Number of ODP faults triggered by RX scatter (e.g., inbound SEND / WRITE operations)" << endl;
    sout << "#odp_tx_res: Number of ODP faults triggered by TX responder gather (e.g., outbound READ RESPONSE)" << endl;
    sout << "#odp_rx_req: Number of ODP faults triggered by RX requestor scatter (e.g., inbound READ RESPONSE)" << endl;

    sout << endl;
}


void IBDiag::DumpDiagnosticCountersP0(
        ofstream &sout,
        struct VS_DiagnosticData *p_p0)
{
    IBDIAG_ENTER;

    struct VS_DC_TransportErrorsAndFlowsV2 * p_p0_v2 =
        &(p_p0->data_set.TransportErrorsAndFlowsV2);

    sout << "#---------- Transport errors and flows (Page0 Rev2)------------"  \
            "----------------" << endl;
    char buffer[2096] = {'\0'};
    snprintf(buffer, sizeof(buffer),
            "rq_num_lle=" U32H_FMT "\n"  \
            "sq_num_lle=" U32H_FMT "\n"  \
            "rq_num_lqpoe=" U32H_FMT "\n"  \
            "sq_num_lqpoe=" U32H_FMT "\n"  \
            "rq_num_leeoe=" U32H_FMT "\n"  \
            "sq_num_leeoe=" U32H_FMT "\n"  \
            "rq_num_lpe=" U32H_FMT "\n"  \
            "sq_num_lpe=" U32H_FMT "\n"  \
            "rq_num_wrfe=" U32H_FMT "\n"  \
            "sq_num_wrfe=" U32H_FMT "\n"  \
            "sq_num_mwbe=" U32H_FMT "\n"  \
            "sq_num_bre=" U32H_FMT "\n"  \
            "rq_num_lae=" U32H_FMT "\n"  \
            "rq_num_rire=" U32H_FMT "\n"  \
            "sq_num_rire=" U32H_FMT "\n"  \
            "rq_num_rae=" U32H_FMT "\n"  \
            "sq_num_rae=" U32H_FMT "\n"  \
            "rq_num_roe=" U32H_FMT "\n"  \
            "sq_num_roe=" U32H_FMT "\n"  \
            "sq_num_rnr=" U32H_FMT "\n"  \
            "rq_num_oos=" U32H_FMT "\n"  \
            "sq_num_oos=" U32H_FMT "\n"  \
            "rq_num_dup=" U32H_FMT "\n"  \
            "sq_num_to=" U32H_FMT "\n"  \
            "sq_num_tree=" U32H_FMT "\n"  \
            "sq_num_rree=" U32H_FMT "\n"  \
            "sq_num_rabrte=" U32H_FMT "\n"  \
            "rq_num_mce=" U32H_FMT "\n"  \
            "rq_num_retrans_rsync=" U32H_FMT "\n"  \
            "sq_num_retrans_rsync=" U32H_FMT "\n"  \
            "sq_num_ldb_drops=" U32H_FMT "\n",
            p_p0_v2->rq_num_lle,
            p_p0_v2->sq_num_lle,
            p_p0_v2->rq_num_lqpoe,
            p_p0_v2->sq_num_lqpoe,
            p_p0_v2->rq_num_leeoe,
            p_p0_v2->sq_num_leeoe,
            p_p0_v2->rq_num_lpe,
            p_p0_v2->sq_num_lpe,
            p_p0_v2->rq_num_wrfe,
            p_p0_v2->sq_num_wrfe,
            p_p0_v2->sq_num_mwbe,
            p_p0_v2->sq_num_bre,
            p_p0_v2->rq_num_lae,
            p_p0_v2->rq_num_rire,
            p_p0_v2->sq_num_rire,
            p_p0_v2->rq_num_rae,
            p_p0_v2->sq_num_rae,
            p_p0_v2->rq_num_roe,
            p_p0_v2->sq_num_roe,
            p_p0_v2->sq_num_rnr,
            p_p0_v2->rq_num_oos,
            p_p0_v2->sq_num_oos,
            p_p0_v2->rq_num_dup,
            p_p0_v2->sq_num_to,
            p_p0_v2->sq_num_tree,
            p_p0_v2->sq_num_rree,
            p_p0_v2->sq_num_rabrte,
            p_p0_v2->rq_num_mce,
            p_p0_v2->rq_num_retrans_rsync,
            p_p0_v2->sq_num_retrans_rsync,
            p_p0_v2->sq_num_ldb_drops);
        sout << buffer;
}


void IBDiag::DumpDiagnosticCountersP1(
        ofstream &sout,
        struct VS_DiagnosticData *p_p1)
{
    IBDIAG_ENTER;

    struct VS_DC_Page1LatestVersion * p_p1_last_ver =
        &(p_p1->data_set.Page1LatestVersion);

    char buffer[2096] = {'\0'};
    sout << "#---------- HCA Extended Flows (Page1  Rev" << PAGE1_LATEST_VER << ")------------"  \
            "----------------" << endl;
    snprintf(buffer, sizeof(buffer),
            "rq_num_sig_err=" U32H_FMT "\n"  \
            "sq_num_sig_err=" U32H_FMT "\n"  \
            "sq_num_cnak=" U32H_FMT "\n"  \
            "sq_reconnect=" U32H_FMT "\n"  \
            "sq_reconnect_ack=" U32H_FMT "\n"  \
            "rq_open_gb=" U32H_FMT "\n"  \
            "rq_num_no_dcrs=" U32H_FMT "\n"  \
            "rq_num_cnak_sent=" U32H_FMT "\n"  \
            "sq_reconnect_ack_bad=" U32H_FMT "\n"  \
            "rq_open_gb_cnak=" U32H_FMT "\n"  \
            "rq_gb_trap_cnak=" U32H_FMT "\n"  \
            "rq_not_gb_connect=" U32H_FMT "\n"  \
            "rq_not_gb_reconnect=" U32H_FMT "\n"  \
            "rq_curr_gb_connect=" U32H_FMT "\n"  \
            "rq_curr_gb_reconnect=" U32H_FMT "\n"  \
            "rq_close_non_gb_gc=" U32H_FMT "\n"  \
            "rq_dcr_inhale_events=" U32H_FMT "\n"  \
            "rq_state_active_gb=" U32H_FMT "\n"  \
            "rq_state_avail_dcrs=" U32H_FMT "\n"  \
            "rq_state_dcr_lifo_size=" U32H_FMT "\n"  \
            "sq_cnak_drop=" U32H_FMT "\n"  \
            "minimum_dcrs=" U32H_FMT "\n"  \
            "maximum_dcrs=" U32H_FMT "\n"  \
            "max_cnak_fifo_size=" U32H_FMT "\n"  \
            "rq_num_dc_cacks=" U32H_FMT "\n"  \
            "sq_num_dc_cacks=" U32H_FMT "\n"
            "rx_rsp_ooo_packet="               U32H_FMT "\n" \
            "rx_req_ooo_packet="               U32H_FMT "\n" \
            "tx_reached_max_outstanding_psn="  U32H_FMT "\n" \
            "tx_pause_on_strong_ordering="     U32H_FMT "\n" \
            // ODP counters
            "odp_fault_discard_eq_overrun="    U32H_FMT "\n" \
            "odp_fault_discard_other="         U32H_FMT "\n" \
            "odp_event_posted="                U32H_FMT "\n" \
            "odp_event_last_posted="           U32H_FMT "\n" \
            "odp_resume_received="             U32H_FMT "\n" \
            "odp_resume_error_received="       U32H_FMT "\n" \
            "odp_tx_req="                      U32H_FMT "\n" \
            "odp_rx_res="                      U32H_FMT "\n" \
            "odp_tx_res="                      U32H_FMT "\n" \
            "odp_rx_req="                      U32H_FMT "\n",

            p_p1_last_ver->rq_num_sig_err,
            p_p1_last_ver->sq_num_sig_err,
            p_p1_last_ver->sq_num_cnak,
            p_p1_last_ver->sq_reconnect,
            p_p1_last_ver->sq_reconnect_ack,
            p_p1_last_ver->rq_open_gb,
            p_p1_last_ver->rq_num_no_dcrs,
            p_p1_last_ver->rq_num_cnak_sent,
            p_p1_last_ver->sq_reconnect_ack_bad,
            p_p1_last_ver->rq_open_gb_cnak,
            p_p1_last_ver->rq_gb_trap_cnak,
            p_p1_last_ver->rq_not_gb_connect,
            p_p1_last_ver->rq_not_gb_reconnect,
            p_p1_last_ver->rq_curr_gb_connect,
            p_p1_last_ver->rq_curr_gb_reconnect,
            p_p1_last_ver->rq_close_non_gb_gc,
            p_p1_last_ver->rq_dcr_inhale_events,
            p_p1_last_ver->rq_state_active_gb,
            p_p1_last_ver->rq_state_avail_dcrs,
            p_p1_last_ver->rq_state_dcr_lifo_size,
            p_p1_last_ver->sq_cnak_drop,
            p_p1_last_ver->minimum_dcrs,
            p_p1_last_ver->maximum_dcrs,
            p_p1_last_ver->max_cnak_fifo_size,
            p_p1_last_ver->rq_num_dc_cacks,
            p_p1_last_ver->sq_num_dc_cacks,
            p_p1_last_ver->rx_rsp_ooo_packet,
            p_p1_last_ver->rx_req_ooo_packet,
            p_p1_last_ver->tx_reached_max_outstanding_psn,
            p_p1_last_ver->tx_pause_on_strong_ordering,
            // ODP counters
            p_p1_last_ver->odp_fault_discard_eq_overrun,
            p_p1_last_ver->odp_fault_discard_other,
            p_p1_last_ver->odp_event_posted,
            p_p1_last_ver->odp_event_last_posted,
            p_p1_last_ver->odp_resume_received,
            p_p1_last_ver->odp_resume_error_received,
            p_p1_last_ver->odp_tx_req,
            p_p1_last_ver->odp_rx_res,
            p_p1_last_ver->odp_tx_res,
            p_p1_last_ver->odp_rx_req);

    sout << buffer;
}

void IBDiag::DumpDiagnosticCountersP255(
        ofstream &sout,
        struct VS_DiagnosticData *p_p255)
{
    IBDIAG_ENTER;

    struct VS_DC_Page255LatestVersion * p_p255_data =
        &(p_p255->data_set.Page255LatestVersion);

    char buffer[2096] = {'\0'};
    sout << "#---------- Internal HCA Debug (Page 255  Rev3) ------------"  \
            "----------------" << endl;

    snprintf(buffer, sizeof(buffer),
            "field0=" U32H_FMT "\n"  \
            "field1=" U32H_FMT "\n"  \
            "field2=" U32H_FMT "\n"  \
            "field3=" U32H_FMT "\n"  \
            "field4=" U32H_FMT "\n"  \
            "field5=" U32H_FMT "\n"  \
            "field6=" U32H_FMT "\n"  \
            "field7=" U32H_FMT "\n"  \
            "field8=" U32H_FMT "\n"  \
            "field9=" U32H_FMT "\n"  \
            "field10=" U32H_FMT "\n"  \
            "field11=" U32H_FMT "\n"  \
            "field12=" U32H_FMT "\n"  \
            "field13=" U32H_FMT "\n"  \
            "field14=" U32H_FMT "\n"  \
            "field15=" U32H_FMT "\n"  \
            "field16=" U32H_FMT "\n",
            p_p255_data->num_cqovf,
            p_p255_data->num_eqovf,
            p_p255_data->sq_num_ds_ex_pi_retry,
            p_p255_data->rq_next_gb_connect,
            p_p255_data->rq_next_gb_reconnect,
            p_p255_data->rq_psn_close_gb,
            p_p255_data->rq_psn_close_gb_cwc,
            p_p255_data->rq_close_gb_gc,
            p_p255_data->rq_cwc_ghost_discard_connect,
            p_p255_data->rq_cwc_ghost_discard_reconnect,
            p_p255_data->rq_cwc_reconnect_stale_connection,
            p_p255_data->rq_cwc_reconnect_probable_current_connection,
            p_p255_data->rq_cwc_uncertainty_discard,
            p_p255_data->rq_num_dc_cacks_full_hash,
            p_p255_data->rq_num_dc_cacks_collision,
            p_p255_data->dc_hash_curr_used_lines,
            p_p255_data->dc_has_max_used_entries_in_line);
    sout << buffer << endl;
}

void IBDiag::DumpDiagnosticCounters(ofstream &sout)
{
    IBDIAG_ENTER;
    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p0 = NULL;
    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p1 = NULL;
    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p255 = NULL;

    char buffer[2096];

    //add header with counters description
    DumpDiagnosticCountersDescriptionP0(sout);
    DumpDiagnosticCountersDescriptionP1(sout);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize();
         ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        p_curr_mlnx_cntrs_p0 =
            this->fabric_extended_info.getVSDiagnosticCountersPage0(i);
        p_curr_mlnx_cntrs_p1 =
            this->fabric_extended_info.getVSDiagnosticCountersPage1(i);
        p_curr_mlnx_cntrs_p255 =
            this->fabric_extended_info.getVSDiagnosticCountersPage255(i);

        if (!p_curr_mlnx_cntrs_p0 && !p_curr_mlnx_cntrs_p1)    //info not exist
            continue;

        snprintf(buffer, sizeof(buffer),
                 "Port=" U32D_FMT    \
                 " Lid=" U16H_FMT    \
                 " Node GUID=" U64H_FMT   \
                 " Port GUID=" U64H_FMT   \
                 " Device=" U32D_FMT \
                 " Port Name=" STR_FMT,
                 p_curr_port->num,
                 p_curr_port->base_lid,
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->p_node->devId,
                 p_curr_port->getName().c_str());
        sout << "-------------------------------------------------------" <<
            endl;
        sout << buffer << endl;
        sout << "-------------------------------------------------------" <<
            endl;


        if (p_curr_mlnx_cntrs_p0)
            DumpDiagnosticCountersP0(sout, p_curr_mlnx_cntrs_p0);

        if (p_curr_mlnx_cntrs_p1)
            DumpDiagnosticCountersP1(sout, p_curr_mlnx_cntrs_p1);

        if (p_curr_mlnx_cntrs_p255)
            DumpDiagnosticCountersP255(sout, p_curr_mlnx_cntrs_p255);
    }
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: IBDIAG_SUCCESS_CODE / IBDIAG_ERR_CODE_FABRIC_ERROR
int IBDiag::CheckVSGeneralInfo(IBNode *p_curr_node, struct VendorSpec_GeneralInfo *p_curr_general_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (p_curr_general_info->FWInfo.Year < 0x2000 ||
        p_curr_general_info->FWInfo.Year > 0x2050 ||
        p_curr_general_info->FWInfo.Day == 0    ||
        p_curr_general_info->FWInfo.Day  > 0x31   ||
        p_curr_general_info->FWInfo.Month == 0  ||
        p_curr_general_info->FWInfo.Month > 0x12) {

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s ,device=%u has invalid FW date %x/%x/%x \n",
                p_curr_node->getName().c_str(),
                p_curr_node->devId,
                p_curr_general_info->FWInfo.Month,
                p_curr_general_info->FWInfo.Day,
                p_curr_general_info->FWInfo.Year);
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpVSNodeInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_NODES_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
            << "HWInfo_DeviceID," << "HWInfo_DeviceHWRevision,"
            << "HWInfo_technology," << "HWInfo_UpTime,"
            << "FWInfo_SubMinor," << "FWInfo_Minor," << "FWInfo_Major,"
            << "FWInfo_secure_fw," << "FWInfo_signed_fw," << "FWInfo_debug_fw," << "FWInfo_dev_fw,"
            << "FWInfo_BuildID,"
            << "FWInfo_Year," << "FWInfo_Day," << "FWInfo_Month," << "FWInfo_Hour,"
            << "FWInfo_PSID,"
            << "FWInfo_INI_File_Version,"
            << "FWInfo_Extended_Major," << "FWInfo_Extended_Minor," << "FWInfo_Extended_SubMinor,"
            << "FWInfo_isfu_major," <<"FWInfo_sec_boot," << "FWInfo_life_cycle,"
            << "SWInfo_SubMinor," << "SWInfo_Minor," << "SWInfo_Major";

    for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS; capability_field++)
         sstream << ",CapabilityMask_" << capability_field;
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    struct VendorSpec_GeneralInfo *p_curr_general_info = NULL;
    char buffer[2096];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        capability_mask_t mask;
        int rc = this->capability_module.GetCapability(p_curr_node, true, mask);
        p_curr_general_info = this->fabric_extended_info.getVSGeneralInfo(i);
        if ((rc != IBDIAG_SUCCESS_CODE) && !p_curr_general_info) //info not exist
            continue;

        sstream.str("");
        sstream << "0x" << std::hex << std::setfill('0') << setw(16)
                << p_curr_node->guid_get() << ',';
        if (p_curr_general_info) {
            string psid = (char *)p_curr_general_info->FWInfo.PSID.PSID;
            snprintf(buffer, sizeof(buffer),
                     U16H_FMT "," U16H_FMT "," U32D_FMT "," U32H_FMT ","  \
                     U8H_FMT "," U8H_FMT "," U8H_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                     U32H_FMT ","  \
                     U16H_FMT "," U8H_FMT "," U8H_FMT "," U16H_FMT ","   \
                     STR_FMT "," \
                     U32H_FMT ","  \
                     U32H_FMT "," U32H_FMT "," U32H_FMT ","  \
                     U16H_FMT "," U32D_FMT "," U32D_FMT "," \
                     U8H_FMT "," U8H_FMT "," U8H_FMT,
                     p_curr_general_info->HWInfo.DeviceID,
                     p_curr_general_info->HWInfo.DeviceHWRevision,
                     p_curr_general_info->HWInfo.technology,
                     p_curr_general_info->HWInfo.UpTime,
                     p_curr_general_info->FWInfo.SubMinor,
                     p_curr_general_info->FWInfo.Minor,
                     p_curr_general_info->FWInfo.Major,
                     p_curr_general_info->FWInfo.secure_fw,
                     p_curr_general_info->FWInfo.signed_fw,
                     p_curr_general_info->FWInfo.debug_fw,
                     p_curr_general_info->FWInfo.dev_fw,
                     p_curr_general_info->FWInfo.BuildID,
                     p_curr_general_info->FWInfo.Year,
                     p_curr_general_info->FWInfo.Day,
                     p_curr_general_info->FWInfo.Month,
                     p_curr_general_info->FWInfo.Hour,
                     (psid == "") ? "UNKNOWN" : psid.c_str(),
                     p_curr_general_info->FWInfo.INI_File_Version,
                     p_curr_general_info->FWInfo.Extended_Major,
                     p_curr_general_info->FWInfo.Extended_Minor,
                     p_curr_general_info->FWInfo.Extended_SubMinor,
                     p_curr_general_info->FWInfo.isfu_major,
                     p_curr_general_info->FWInfo.sec_boot,
                     p_curr_general_info->FWInfo.life_cycle,
                     p_curr_general_info->SWInfo.SubMinor,
                     p_curr_general_info->SWInfo.Minor,
                     p_curr_general_info->SWInfo.Major);
            sstream << buffer;
        } else {
            sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                    << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                    << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A";
        }

        if (rc == IBDIAG_SUCCESS_CODE)
            for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS;
                  capability_field++)
                  sstream << ",0x" << setw(8) << mask.mask[capability_field];
        else
             for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS;
                  capability_field++)
                  sstream << ",N/A";

        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_NODES_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteNodesInfoFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Nodes Information",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpNodesInfo(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_CHECK_FAILED
int IBDiag::CheckFWVersion(list_p_fabric_general_err& fw_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    map_devid_psid_to_fw_version_obj_t max_fw_version_per_devid_psid;

    fw_version_obj_t curr_devid_fw_version;
    fw_version_obj_t max_devid_fw_version;

    //find maximum fw version for each device id
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        VendorSpec_GeneralInfo *p_curr_general_info = this->fabric_extended_info.getVSGeneralInfo(i);
        if (!p_curr_general_info)           //info not exists
            continue;

        int rc2 = this->CheckVSGeneralInfo(p_curr_node, p_curr_general_info);
        if (rc2) {
            //mark this as invalid data
            p_curr_general_info->FWInfo.Day=0;
            continue;
        }

        curr_devid_fw_version = *p_curr_general_info;
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s has FW Version "    \
                U32D_FMT "." U32D_FMT "." U32D_FMT    \
                " ,device=%u\n",
                p_curr_node->getName().c_str(),
                curr_devid_fw_version.major, curr_devid_fw_version.minor, curr_devid_fw_version.sub_minor,
                p_curr_node->devId);

        string psid_string((char *)p_curr_general_info->FWInfo.PSID.PSID, 16);

        map_devid_psid_to_fw_version_obj_t::iterator it2 =
           max_fw_version_per_devid_psid.find(pair_device_id_t_string(
                       p_curr_node->devId,
                       psid_string));

        if (it2 == max_fw_version_per_devid_psid.end()) {      //first time reached this devid
            max_fw_version_per_devid_psid[pair_device_id_t_string(
                    p_curr_node->devId, psid_string)] =
                p_curr_general_info;
            continue;
        }
        max_devid_fw_version = *(it2->second);
        if (curr_devid_fw_version > max_devid_fw_version) {
            max_fw_version_per_devid_psid[pair_device_id_t_string(
                    p_curr_node->devId,
                    psid_string)] =
                p_curr_general_info;
        }
    }
    for (map_devid_psid_to_fw_version_obj_t::iterator it =
            max_fw_version_per_devid_psid.begin();
         it != max_fw_version_per_devid_psid.end();
         it++) {
        pair_device_id_t_string pair_devid_psid = (it->first);
        curr_devid_fw_version = *(it->second);
        INFO_PRINT("Devid: " U32D_FMT "(" U16H_FMT "), PSID: " STR_FMT ", Latest FW Version:"    \
                   U32D_FMT "." U32D_FMT "." U32D_FMT "\n",
                   pair_devid_psid.first,
                   pair_devid_psid.first,
                   pair_devid_psid.second.c_str(),
                   curr_devid_fw_version.major,
                   curr_devid_fw_version.minor,
                   curr_devid_fw_version.sub_minor);
    }

    //find nodes with not maximum fw version
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        VendorSpec_GeneralInfo *p_curr_general_info = this->fabric_extended_info.getVSGeneralInfo(i);
        if (!p_curr_general_info)           //info not exists
            continue;
        if (p_curr_general_info->FWInfo.Day==0) {
            fw_errors.push_back(
                new FabricErrNodeNotSupportCap(p_curr_node,
                        "The firmware of this device returned invalid general info data"));

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            continue;
        }
        curr_devid_fw_version = *p_curr_general_info;
        string psid_string((char *)p_curr_general_info->FWInfo.PSID.PSID, 16);

        map_devid_psid_to_fw_version_obj_t::iterator it =
            max_fw_version_per_devid_psid.find(pair_device_id_t_string(
                        p_curr_node->devId,
                        psid_string));
        if (it == max_fw_version_per_devid_psid.end()) {      //first time reached this devid
            this->SetLastError("DB Error - failed to find maximum FW version for node=%s, device=%u",
                    p_curr_node->getName().c_str(), p_curr_node->devId);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto check_fw_version_exit;
        }
        max_devid_fw_version = *(it->second);
        if (curr_devid_fw_version != max_devid_fw_version) {
            char buffer[1024];
            snprintf(buffer, sizeof(buffer),
                     "Node with Devid:" U32D_FMT "(" U16H_FMT ")"    \
                     ",PSID:" STR_FMT " has FW version " \
                     U32D_FMT "." U32D_FMT "." U32D_FMT  \
                     " while the latest FW version for the same Devid/PSID on this fabric is "   \
                     U32D_FMT "." U32D_FMT "." U32D_FMT,
                     p_curr_node->devId, p_curr_node->devId,
                     psid_string.c_str(),
                     curr_devid_fw_version.major, curr_devid_fw_version.minor, curr_devid_fw_version.sub_minor,
                     max_devid_fw_version.major, max_devid_fw_version.minor, max_devid_fw_version.sub_minor);

            FabricErrNodeWrongFWVer *p_curr_fabric_node_err =
                    new FabricErrNodeWrongFWVer(p_curr_node, buffer);

            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            fw_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

check_fw_version_exit:
    max_fw_version_per_devid_psid.clear();
    IBDIAG_RETURN(rc);
}


/*****************************************************************************/
/*       Vendor Specific Capability SMP                                      */
/*****************************************************************************/

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVsCapSmpFwInfo(list_p_fabric_general_err &vs_cap_smp_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPVSGeneralInfoFwInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect FW
    struct FWInfo_Block_Element curr_general_info;
    CLEAR_STRUCT(curr_general_info);
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());

            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //currently we will send SMP FW MAD only for acquiring
        //SMP capability mask purposes
        //change it once, we need SMP FW for other purposes
        if (capability_module.IsSMPMaskKnown(p_curr_node->guid_get()))
            continue;

        //send general info mad get for this node, if it allows it (i.e supports such mad)
        capability_mask_t mask;
        if (capability_module.IsSMPUnsupportedMadDevice(p_curr_node->vendId,
                                                        p_curr_node->devId,
                                                        mask))
            continue;
        clbck_data.m_data1 = p_curr_node;
        direct_route_t *p_direct_route = this->GetDR(p_curr_node);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        progress_bar.push(p_curr_node);
        this->ibis_obj.SMPVSGeneralInfoFwInfoMadGetByDirect(p_direct_route,
                                                            &curr_general_info,
                                                            &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!vs_cap_smp_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildVsCapSmpCapabilityMask(list_p_fabric_general_err &vs_cap_smp_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPVSGeneralInfoCapabilityMaskGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    bool query_per_matched_guid;

    // go over all nodes and collect Capability Mask
    struct GeneralInfoCapabilityMask curr_general_info;
    CLEAR_STRUCT(curr_general_info);
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName"
                               " map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        //check if this node has already known mask, i.e in the cache
        //we already took care of fw configured masks in BuildVsCapSmpFwInfo,
        //so the mask should be known in that case
        if (capability_module.IsSMPMaskKnown(p_curr_node->guid_get()))
            continue;

        //check if there is configuration per GUID that says query
        query_per_matched_guid = false;
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;

        //these checks are to prevent programming errors
        //as we already took care of 'unsupported devices', 'guid to mask' and
        //'ven id, dev id, fw, mask' configurations
        //so now only 'guid query', 'ven id dev id fw query' are left
        if (capability_module.IsLongestSMPPrefixMatch(p_curr_node->guid_get(),
                                                      prefix_len,
                                                      matched_guid,
                                                      qmask) &&
            qmask.to_query == true)
            query_per_matched_guid = true;

        capability_mask_t mask;

        mask.clear();

        //no match by guid with 'query'
        if (!query_per_matched_guid) {
            //check if it matches unsupported device then
            if (capability_module.IsSMPUnsupportedMadDevice(p_curr_node->vendId,
                                                            p_curr_node->devId,
                                                            mask))
                continue;
        }

        clbck_data.m_data1 = p_curr_node;
        direct_route_t *p_direct_route = this->GetDR(p_curr_node);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        progress_bar.push(p_curr_node);
        this->ibis_obj.SMPVSGeneralInfoCapabilityMaskMadGetByDirect(
            p_direct_route, &curr_general_info, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!vs_cap_smp_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::ValidateExtendedNodeInfoForSwitches()
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    struct ib_extended_node_info *p_meni;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
         nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_node = *nI;
        if (!p_node) {
           this->SetLastError("DB error - found null node in Switches set");
           IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        p_meni = this->fabric_extended_info.getSMPExtNodeInfo(p_node->createIndex);
        if (!p_meni)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildExtendedNodeInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPExtendedNodeInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;

    // go over all nodes:
    struct ib_extended_node_info ext_node_info;
    CLEAR_STRUCT(ext_node_info);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        if (!this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsExtendedNodeInfoSupported))
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.SMPExtendedNodeInfoMadGetSetByDirect(p_curr_direct_route,
                                                            IBIS_IB_MAD_METHOD_GET,
                                                            &ext_node_info,
                                                            &clbck_data);

        if (ibDiagClbck.GetState())
            break;

    } // nodes loop

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildExtendedPortInfo(list_p_fabric_general_err &vs_ext_pi_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    ibDiagClbck.Set(this, &fabric_extended_info, &vs_ext_pi_errors,
                    NULL, &capability_module);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPVSExtendedPortInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect extended port info
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // go over all ports of this node:
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) { //start from port 1, no need to reset port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;
            SMP_PortInfo *curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!curr_port_info) {
                this->SetLastError("DB error - found connected port=%s"
                                   " without SMPPortInfo",
                                   p_curr_port->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            /* check for MLNX speed if needed and LLR active cell size
               if we had extended speed than we sure not QDR and
               we do not need to check for MLNX speed */
            /*
             * Check if need to send MAD
             * to query for mlnx extended speeds and LLR Active Cell size
             */

            if (curr_port_info->PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)
                continue;
            //is this device supports Mlnx Extended Port Info mad
            bool dev_support_mepi = this->capability_module.IsSupportedSMPCapability(
                p_curr_port->p_node, EnSMPCapIsExtendedPortInfoSupported);

            //Don't send the mad to device that don't support it
            if (!dev_support_mepi) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                           "The port %s doesn't support VS"
                           " SMPExtendedPortInfo MAD\n",
                           p_curr_port->getName().c_str());
                continue;
            }


            SMP_MlnxExtPortInfo curr_mlnx_ext_port_info;

            /*
             * Send the MAD
             */
            clbck_data.m_data1 = p_curr_port;

            direct_route_t * p_direct_route = this->GetDR(p_curr_port->p_node);
            if (!p_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s",
                                   p_curr_node->getName().c_str());

                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            progress_bar.push(p_curr_port);
            this->ibis_obj.SMPMlnxExtPortInfoMadGetByDirect(
                p_direct_route,
                p_curr_port->num,
                &curr_mlnx_ext_port_info,
                &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS ExtendedPortInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of VS ExtendedPortInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!vs_ext_pi_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

/*****************************************************************************/
/*       Vendor Specific Capability GMP                                      */
/*****************************************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVsCapGmpInfo(list_p_fabric_general_err &vs_cap_gmp_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::VSGeneralInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    bool query_per_matched_guid;

    // go over all nodes and collect FW
    struct VendorSpec_GeneralInfo curr_general_info;
    CLEAR_STRUCT(curr_general_info);
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        IBPort *p_curr_port = NULL;
        if(p_curr_node->type == IB_SW_NODE) {
            p_curr_port = p_curr_node->getPort(0);  //getting port 0
            if (!p_curr_port || p_curr_port->base_lid == 0 || p_curr_port->base_lid >= 0xc000)   {   //port not exist
                continue;
            }
        } else {
            bool found = false;
            for (unsigned int i = 1 ; i <=  p_curr_node->numPorts && !found; i++) {
                p_curr_port = p_curr_node->getPort((phys_port_t)i);
                found = p_curr_port && p_curr_port->base_lid && p_curr_port->base_lid < 0xc000;
            }
            if (!found) //all ports not exist
                continue;

            if (p_curr_node->isSpecialNode())
                continue;
        }

        //send general info mad get for this node, if it supports the mad
        capability_mask_t mask;
        mask.clear();

        //check if there is configuration per GUID that says query
        query_per_matched_guid = false;
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;

        //don't send the mad for devices that don't support it
        //however the capability configuration 'guid -> query', precedes
        //this rule
        if (capability_module.IsLongestGMPPrefixMatch(p_curr_node->guid_get(),
                                                      prefix_len,
                                                      matched_guid,
                                                      qmask) &&
            qmask.to_query == true)
            query_per_matched_guid = true;

        if (!query_per_matched_guid &&
            capability_module.IsGMPUnsupportedMadDevice(p_curr_node->vendId,
                                                        p_curr_node->devId,
                                                        mask))
                continue;

        //send mad anyway for Nodes Info file
        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);
        this->ibis_obj.VSGeneralInfoGet(p_curr_port->base_lid,
                                        &curr_general_info, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!vs_cap_gmp_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVsCapSmp(list_p_fabric_general_err &vs_cap_smp_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc1, rc2;

    ibDiagClbck.Set(this, &fabric_extended_info, &vs_cap_smp_errors, NULL, &capability_module);
    //retrieve fw info
    INFO_PRINT("Build VS Capability FW Info SMP\n");
    rc1 = this->BuildVsCapSmpFwInfo(vs_cap_smp_errors);
    SCREEN_PRINT("\n");

    //retrieve capability mask
    INFO_PRINT("Build VS Capability Mask SMP\n");
    rc2 = this->BuildVsCapSmpCapabilityMask(vs_cap_smp_errors);

    IBDIAG_RETURN((rc1 || rc2) ? IBDIAG_ERR_CODE_FABRIC_ERROR : IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVsCapGmp(list_p_fabric_general_err &vs_cap_gmp_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    ibDiagClbck.Set(this, &fabric_extended_info, &vs_cap_gmp_errors,
                    NULL, &capability_module);
    //retrieve fw info
    rc = this->BuildVsCapGmpInfo(vs_cap_gmp_errors);

    IBDIAG_RETURN(rc);
}

/*****************************************************************************/
/*       Vendor Specific Diagnostic Counters                                 */
/*****************************************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR /
//ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVsDiagnosticCounters(
        list_p_fabric_general_err& mlnx_cntrs_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &mlnx_cntrs_errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect counters
    struct VS_DiagnosticData p_mlnx_cntrs;
    CLEAR_STRUCT(p_mlnx_cntrs);
    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        IBPort *p_curr_port = NULL;
        if(p_curr_node->type != IB_SW_NODE) {
             struct SMP_NodeInfo * p_curr_node_info =
                    this->fabric_extended_info.getSMPNodeInfo(
                            p_curr_node->createIndex);
             if (!p_curr_node_info) {
                 this->SetLastError(
                         "DB error - failed to get SMPNodeInfo for node: %s",
                         p_curr_node->name.c_str());
                 rc = IBDIAG_ERR_CODE_DB_ERR;
                 goto exit;
             }

            if (!this->capability_module.IsSupportedGMPCapability(
                        p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                           "node %s: DeviceID %u (0x%x) Does not support"  \
                           " Diagnostic Counters, skipping \n",
                        p_curr_node->name.c_str(),
                        p_curr_node_info->DeviceID,
                        p_curr_node_info->DeviceID);
                continue;
            }

            for (unsigned int i = 1 ; i <=  p_curr_node->numPorts ; i++) {
                p_curr_port = p_curr_node->getPort((phys_port_t)i);
                //Skip ports that don't exist or are down or have undefined
                //state or are not in the scope of the fabric.
                if (!p_curr_port || !p_curr_port->is_data_worthy())
                    continue;

                clbck_data.m_data1 = p_curr_port;
                //get counters per node, so the port number is irrelevant, i.e 0
                //page 0
                progress_bar.push(p_curr_port);
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::VSDiagnosticCountersPage0GetClbck>;

                this->ibis_obj.VSDiagnosticDataGet(
                        p_curr_port->base_lid,
                        0,
                        VS_MLNX_CNTRS_PAGE0,
                        &p_mlnx_cntrs,
                        &clbck_data);

                //page 1
                progress_bar.push(p_curr_port);
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::VSDiagnosticCountersPage1GetClbck>;

                this->ibis_obj.VSDiagnosticDataGet(
                        p_curr_port->base_lid,
                        0,
                        VS_MLNX_CNTRS_PAGE1,
                        &p_mlnx_cntrs,
                        &clbck_data);

                //page 255
                progress_bar.push(p_curr_port);
                clbck_data.m_handle_data_func =
                    &forwardClbck<IBDiagClbck, &IBDiagClbck::VSDiagnosticCountersPage255GetClbck>;

                this->ibis_obj.VSDiagnosticDataGet(
                        p_curr_port->base_lid,
                        0,
                        VS_MLNX_CNTRS_PAGE255,
                        &p_mlnx_cntrs,
                        &clbck_data);

                //the information is node related, so one port is enough
                //when will add info per port, add NOT_SUPPORT_MLNX_COUNTERS
               break;
            }
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!mlnx_cntrs_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR /
//ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::ResetDiagnosticCounters(
        list_p_fabric_general_err& mlnx_cntrs_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &mlnx_cntrs_errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect FW
    struct VS_DiagnosticData p_mlnx_cntrs;
    CLEAR_STRUCT(p_mlnx_cntrs);
    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        if (!p_curr_node->getInSubFabric() || p_curr_node->type == IB_SW_NODE)
            continue;

        IBPort *p_curr_port = NULL;
        struct SMP_NodeInfo * p_curr_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError(
                    "DB error - failed to get SMPNodeInfo for node: %s",
                    p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!this->capability_module.IsSupportedGMPCapability(
                    p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s: DeviceID %u (0x%x)"   \
                                           " Does not support Mellanox"  \
                                           " Counters, skipping \n",
                    p_curr_node->name.c_str(),
                    p_curr_node_info->DeviceID,
                    p_curr_node_info->DeviceID);
            continue;
        }

        for (unsigned int i = 1 ; i <=  p_curr_node->numPorts ; i++) {
            p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_port;
            // Clear all the diagnostic pages: pages 0, 1, 255
            // page 0
            clbck_data.m_handle_data_func =
                &forwardClbck<IBDiagClbck, &IBDiagClbck::VSDiagnosticCountersClearClbck>;

            progress_bar.push(p_curr_port);
            this->ibis_obj.VSDiagnosticDataPageClear(
                    p_curr_port->base_lid,
                    0, // clear per node, so port number is irrelevant
                    VS_MLNX_CNTRS_PAGE0, &p_mlnx_cntrs,
                    &clbck_data);
            //page 1
            progress_bar.push(p_curr_port);
            this->ibis_obj.VSDiagnosticDataPageClear(
                    p_curr_port->base_lid,
                    0, // clear per node, so port number is irrelevant
                    VS_MLNX_CNTRS_PAGE1, &p_mlnx_cntrs,
                    &clbck_data);
            //page 255
            progress_bar.push(p_curr_port);
            this->ibis_obj.VSDiagnosticDataPageClear(
                    p_curr_port->base_lid,
                    0, // clear per node, so port number is irrelevant
                    VS_MLNX_CNTRS_PAGE255, &p_mlnx_cntrs,
                    &clbck_data);
            //the information is node related, so one port is enough
            break;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!mlnx_cntrs_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpDiagnosticCountersToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p0 = NULL;
    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p1 = NULL;
    struct VS_DiagnosticData *p_curr_mlnx_cntrs_p255 = NULL;
    struct VS_DC_TransportErrorsAndFlowsV2 *p_p0_v2 = NULL;
    struct VS_DC_Page1LatestVersion *p_p1_last_ver = NULL;
    struct VS_DC_Page255LatestVersion *p_p255_last_ver = NULL;

    char buffer[1024] = {'\0'};

    if (csv_out.DumpStart(SECTION_MLNX_CNTRS_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNumber," "LID,";
    // Page0
    sstream << "rq_num_lle," << "sq_num_lle," << "rq_num_lqpoe,"
            << "sq_num_lqpoe," << "rq_num_leeoe," << "sq_num_leeoe,"
            << "rq_num_lpe," << "sq_num_lpe," << "rq_num_wrfe,"
            << "sq_num_wrfe," << "sq_num_mwbe," << "sq_num_bre,"
            << "rq_num_lae," << "rq_num_rire,"
            << "sq_num_rire," << "rq_num_rae,"
            << "sq_num_rae,"
            << "rq_num_roe," << "sq_num_roe,"
            << "sq_num_rnr," << "rq_num_oos,"
            << "sq_num_oos," << "rq_num_dup,"
            << "sq_num_to," << "sq_num_tree," << "sq_num_rree,"
            << "sq_num_rabrte," << "rq_num_mce," << "rq_num_retrans_rsync,"
            << "sq_num_retrans_rsync," << "sq_num_ldb_drops,";
    // Page1
    sstream << "rq_num_sig_err," << "sq_num_sig_err," << "sq_num_cnak,"
            << "sq_reconnect," << "sq_reconnect_ack," << "rq_open_gb,"
            << "rq_num_no_dcrs," << "rq_num_cnak_sent,"
            << "sq_reconnect_ack_bad," << "rq_open_gb_cnak,"
            << "rq_gb_trap_cnak," << "rq_not_gb_connect,"
            << "rq_not_gb_reconnect," << "rq_curr_gb_connect,"
            << "rq_curr_gb_reconnect," << "rq_close_non_gb_gc,"
            << "rq_dcr_inhale_events," << "rq_state_active_gb,"
            << "rq_state_avail_dcrs," << "rq_state_dcr_lifo_size,"
            << "sq_cnak_drop," << "minimum_dcrs," << "maximum_dcrs,"
            << "max_cnak_fifo_size," << "rq_num_dc_cacks," << "sq_num_dc_cacks,"
            << "rx_rsp_ooo_packet," << "rx_req_ooo_packet,"
            << "tx_reached_max_outstanding_psn," << "tx_pause_on_strong_ordering,"
            // ODP counters
            << "odp_fault_discard_eq_overrun," << "odp_fault_discard_other,"
            << "odp_event_posted," << "odp_event_last_posted,"
            << "odp_resume_received," << "odp_resume_error_received,"
            << "odp_tx_req," << "odp_rx_res," << "odp_tx_res," << "odp_rx_req";

    // Page 255
    for (int i = 0; i < PAGE255_NUM_FIELDS; ++i) {
        sstream << ",field" << i;
    }

    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {

        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        p_curr_mlnx_cntrs_p0 =
            this->fabric_extended_info.getVSDiagnosticCountersPage0(i);
        p_curr_mlnx_cntrs_p1 =
            this->fabric_extended_info.getVSDiagnosticCountersPage1(i);
        p_curr_mlnx_cntrs_p255 =
            this->fabric_extended_info.getVSDiagnosticCountersPage255(i);

        if (!p_curr_mlnx_cntrs_p0 && !p_curr_mlnx_cntrs_p1 && !p_curr_mlnx_cntrs_p255)  //info not exist
            continue;

        sstream.str("");
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT "," U16H_FMT ",",
                 p_curr_port->p_node->guid_get(),
                 p_curr_port->guid_get(),
                 p_curr_port->num,
                 p_curr_port->base_lid);

        sstream << buffer;

        if (p_curr_mlnx_cntrs_p0) {
            p_p0_v2 =
                &(p_curr_mlnx_cntrs_p0->data_set.TransportErrorsAndFlowsV2);
            snprintf(buffer, sizeof(buffer),
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                     U32D_FMT ",",
                     p_p0_v2->rq_num_lle,
                     p_p0_v2->sq_num_lle,
                     p_p0_v2->rq_num_lqpoe,
                     p_p0_v2->sq_num_lqpoe,
                     p_p0_v2->rq_num_leeoe,
                     p_p0_v2->sq_num_leeoe,
                     p_p0_v2->rq_num_lpe,
                     p_p0_v2->sq_num_lpe,
                     p_p0_v2->rq_num_wrfe,
                     p_p0_v2->sq_num_wrfe,
                     p_p0_v2->sq_num_mwbe,
                     p_p0_v2->sq_num_bre,
                     p_p0_v2->rq_num_lae,
                     p_p0_v2->rq_num_rire,
                     p_p0_v2->sq_num_rire,
                     p_p0_v2->rq_num_rae,
                     p_p0_v2->sq_num_rae,
                     p_p0_v2->rq_num_roe,
                     p_p0_v2->sq_num_roe,
                     p_p0_v2->sq_num_rnr,
                     p_p0_v2->rq_num_oos,
                     p_p0_v2->sq_num_oos,
                     p_p0_v2->rq_num_dup,
                     p_p0_v2->sq_num_to,
                     p_p0_v2->sq_num_tree,
                     p_p0_v2->sq_num_rree,
                     p_p0_v2->sq_num_rabrte,
                     p_p0_v2->rq_num_mce,
                     p_p0_v2->rq_num_retrans_rsync,
                     p_p0_v2->sq_num_retrans_rsync,
                     p_p0_v2->sq_num_ldb_drops);

        } else {
            snprintf(buffer, sizeof(buffer), "-1"                      \
                     ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1" \
                     ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1" \
                     ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,");
        }
        sstream << buffer;

        // Page 1
        if (p_curr_mlnx_cntrs_p1) {
            p_p1_last_ver = &(p_curr_mlnx_cntrs_p1->data_set.Page1LatestVersion);

            sstream << p_p1_last_ver->rq_num_sig_err << ","
                    << p_p1_last_ver->sq_num_sig_err << ","
                    << p_p1_last_ver->sq_num_cnak << ","
                    << p_p1_last_ver->sq_reconnect << ","
                    << p_p1_last_ver->sq_reconnect_ack << ","
                    << p_p1_last_ver->rq_open_gb << ","
                    << p_p1_last_ver->rq_num_no_dcrs << ","
                    << p_p1_last_ver->rq_num_cnak_sent << ","
                    << p_p1_last_ver->sq_reconnect_ack_bad << ","
                    << p_p1_last_ver->rq_open_gb_cnak << ","
                    << p_p1_last_ver->rq_gb_trap_cnak << ","
                    << p_p1_last_ver->rq_not_gb_connect << ","
                    << p_p1_last_ver->rq_not_gb_reconnect << ","
                    << p_p1_last_ver->rq_curr_gb_connect << ","
                    << p_p1_last_ver->rq_curr_gb_reconnect << ","
                    << p_p1_last_ver->rq_close_non_gb_gc << ","
                    << p_p1_last_ver->rq_dcr_inhale_events << ","
                    << p_p1_last_ver->rq_state_active_gb << ","
                    << p_p1_last_ver->rq_state_avail_dcrs << ","
                    << p_p1_last_ver->rq_state_dcr_lifo_size << ","
                    << p_p1_last_ver->sq_cnak_drop << ","
                    << p_p1_last_ver->minimum_dcrs << ","
                    << p_p1_last_ver->maximum_dcrs << ","
                    << p_p1_last_ver->max_cnak_fifo_size << ","
                    << p_p1_last_ver->rq_num_dc_cacks << ","
                    << p_p1_last_ver->sq_num_dc_cacks << ","
                    << p_p1_last_ver->rx_rsp_ooo_packet               << ","
                    << p_p1_last_ver->rx_req_ooo_packet               << ","
                    << p_p1_last_ver->tx_reached_max_outstanding_psn  << ","
                    << p_p1_last_ver->tx_pause_on_strong_ordering     << ","
                    // ODP counters
                    << p_p1_last_ver->odp_fault_discard_eq_overrun    << ","
                    << p_p1_last_ver->odp_fault_discard_other         << ","
                    << p_p1_last_ver->odp_event_posted                << ","
                    << p_p1_last_ver->odp_event_last_posted           << ","
                    << p_p1_last_ver->odp_resume_received             << ","
                    << p_p1_last_ver->odp_resume_error_received       << ","
                    << p_p1_last_ver->odp_tx_req                      << ","
                    << p_p1_last_ver->odp_rx_res                      << ","
                    << p_p1_last_ver->odp_tx_res                      << ","
                    << p_p1_last_ver->odp_rx_req                      << ",";
        } else {
            sstream << "-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,"
                    << "-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,"
                    // ODP counters
                    << "-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,";
        }

        // Page 255
        if (p_curr_mlnx_cntrs_p255) {
            p_p255_last_ver = &(p_curr_mlnx_cntrs_p255->data_set.Page255LatestVersion);

            sstream << p_p255_last_ver->num_cqovf << ","
                    << p_p255_last_ver->num_eqovf << ","
                    << p_p255_last_ver->sq_num_ds_ex_pi_retry << ","
                    << p_p255_last_ver->rq_next_gb_connect << ","
                    << p_p255_last_ver->rq_next_gb_reconnect << ","
                    << p_p255_last_ver->rq_psn_close_gb << ","
                    << p_p255_last_ver->rq_psn_close_gb_cwc << ","
                    << p_p255_last_ver->rq_close_gb_gc << ","
                    << p_p255_last_ver->rq_cwc_ghost_discard_connect << ","
                    << p_p255_last_ver->rq_cwc_ghost_discard_reconnect << ","
                    << p_p255_last_ver->rq_cwc_reconnect_stale_connection << ","
                    << p_p255_last_ver->rq_cwc_reconnect_probable_current_connection << ","
                    << p_p255_last_ver->rq_cwc_uncertainty_discard << ","
                    << p_p255_last_ver->rq_num_dc_cacks_full_hash << ","
                    << p_p255_last_ver->rq_num_dc_cacks_collision << ","
                    << p_p255_last_ver->dc_hash_curr_used_lines << ","
                    << p_p255_last_ver->dc_has_max_used_entries_in_line;
        } else {
            sstream << "-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1";
        }

        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_MLNX_CNTRS_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteMlnxCntrsFile(const string &file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile("Mellanox Counters",
                            file_name, sout,
                            false, //no append
                            true); //add_header
    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpDiagnosticCounters(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::CheckCapabilityForQoSConfigSL(IBNode *p_curr_node, bool is_vports,
                                          list_p_fabric_general_err& qos_config_sl_errors,
                                          bool& has_capability)
{
    IBDIAG_ENTER;

    EnSMPCapabilityMaskBit_t capability_mask_bit_qos_config_sl_rate_limit;
    EnSMPCapabilityMaskBit_t capability_mask_bit_qos_config_sl_bw;
    string buff_port_type;
    if (is_vports) {
        capability_mask_bit_qos_config_sl_rate_limit = EnSMPCapIsQoSConfigSLVPortRateLimitSupported;
        capability_mask_bit_qos_config_sl_bw = EnSMPCapIsQoSConfigSLVPortAllocBWSupported;
        buff_port_type = " for vports";
    } else {
        capability_mask_bit_qos_config_sl_rate_limit = EnSMPCapIsQoSConfigSLRateLimitSupported;
        capability_mask_bit_qos_config_sl_bw = EnSMPCapIsQoSConfigSLAllocBWSupported;
        buff_port_type = "";
    }

    bool rate_limit_flag = this->capability_module.IsSupportedSMPCapability(
            p_curr_node, capability_mask_bit_qos_config_sl_rate_limit);

    bool bw_alloc_flag = this->capability_module.IsSupportedSMPCapability(
            p_curr_node, capability_mask_bit_qos_config_sl_bw);

    has_capability = true;
    if (!rate_limit_flag && !bw_alloc_flag) {
        struct SMP_NodeInfo * p_curr_node_info =
        this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError("DB error - failed to get SMPNodeInfo for node: %s",
                    p_curr_node->name.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                "node %s: DeviceID %u (0x%x) Does not support"  \
                " QoS Config SL MAD %s, skipping \n",
                p_curr_node->name.c_str(),
                p_curr_node_info->DeviceID,
                p_curr_node_info->DeviceID,
                buff_port_type.c_str());

        has_capability = false;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildSMPQoSConfigSL(list_p_fabric_general_err& qos_config_sl_errors,
                                bool is_vports)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &qos_config_sl_errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    if (is_vports)
        clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::IBDiagSMPVPortQoSConfigSLGetClbck>;
    else
        clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::IBDiagSMPQoSConfigSLGetClbck>;

    struct SMP_QosConfigSL qos_config_sl;
    CLEAR_STRUCT(qos_config_sl);

    // go over all nodes and collect qos_config_sl
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // send this MAD only for CAs.
        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_CA_NODE)
            continue;

        bool has_capability;
        rc = CheckCapabilityForQoSConfigSL(p_curr_node, is_vports, qos_config_sl_errors,
                                           has_capability);
        if (rc)
            goto exit;

        // go over all ports of this node:
        for (u_int32_t i = 0; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);

            if (!p_curr_port)
                continue;

            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_port;

            //get direct route for this port
            direct_route_t *p_curr_direct_route =
                    this->GetDR(p_curr_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s, port=%u",
                                   p_curr_node->getName().c_str(), (u_int8_t)p_curr_port->num);
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            if (has_capability && !is_vports) {
                progress_bar.push(p_curr_port);
                this->ibis_obj.SMPQosConfigSLGetByDirect(p_curr_direct_route,
                                                         &qos_config_sl,
                                                         &clbck_data,
                                                         0); // for CAs - ignore port number
                if(ibDiagClbck.GetState())
                    goto exit;
            }

            if (has_capability && is_vports) {
                map_vportnum_vport vports = p_curr_port->VPorts;
                for (map_vportnum_vport::iterator vpI = vports.begin();
                        vpI != vports.end(); ++vpI) {
                     IBVPort *p_vport = vpI->second;

                     if(!p_vport)
                         continue;

                     clbck_data.m_data2 = p_vport;
                     progress_bar.push(p_curr_port);
                     this->ibis_obj.SMPVPortQoSConfigSLGetByDirect(p_curr_direct_route,
                                                                   &qos_config_sl,
                                                                   &clbck_data,
                                                                   p_vport->getVPortNum());

                     if(ibDiagClbck.GetState())
                        goto exit;
                }
            }
        } // ports loop
    } // nodes loop

exit:

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!qos_config_sl_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpQoSConfigSLToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_QOS_CONFIG_SL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNum,SL,BandwidthShare,RateLimit" << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_QosConfigSL *p_curr_qos_config_sl = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        if (p_curr_node->type != IB_CA_NODE)
            continue;

        bool rate_limit_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLRateLimitSupported);

        bool bw_alloc_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLAllocBWSupported);

        // go over all ports of this node:
        for (u_int32_t pI = 0; pI <= p_curr_node->numPorts; ++pI) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)pI);

            if (!p_curr_port)
                continue;
            if (!p_curr_port->is_data_worthy())
                continue;

            p_curr_qos_config_sl = fabric_extended_info.getSMPQosConfigSL(p_curr_port->createIndex);
            if (!p_curr_qos_config_sl)           //info not exist
                continue;

            for (u_int32_t slI = 0; slI < IB_NUM_SL; ++slI) {
                sstream.str("");
                snprintf(buffer, sizeof(buffer), U64H_FMT "," U64H_FMT",%d,%d,",
                         p_curr_node->guid_get(),
                         p_curr_port->guid_get(),
                         p_curr_port->num,
                         slI);

                sstream << buffer;
                bw_alloc_flag? sstream << p_curr_qos_config_sl->BandwidthPerSL[slI].BandwidthShare
                        : sstream << "N/A";
                sstream << ",";
                rate_limit_flag? sstream << p_curr_qos_config_sl->BandwidthPerSL[slI].RateLimit
                        : sstream << "N/A";
                sstream << endl;

                csv_out.WriteBuf(sstream.str());
            }
        }
    }

    csv_out.DumpEnd(SECTION_QOS_CONFIG_SL);
    IBDIAG_RETURN(rc);
}

int IBDiag::BuildSMPQoSConfigVL(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::IBDiagSMPQoSConfigVLGetClbck>;

    struct SMP_QosConfigVL qos_config_vl = {};

    for (auto node_pair : this->discovered_fabric.NodeByName) {

        IBNode *p_node = node_pair.second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_node->getInSubFabric())
            continue;

        ib_extended_node_info* p_eni =
                this->fabric_extended_info.getSMPExtNodeInfo(p_node->createIndex);

        if (!p_eni)
            continue;

        // send only to Switches and GPUs
        if (p_node->type != IB_SW_NODE && p_eni->node_type_extended != IB_GPU_NODE)
            continue;

        // check if node support QoSConfigVL
        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                        p_node, EnSMPCapIsQoSConfigVLSupported))
            continue;

        // go over all ports of this node:
        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);

            if (!p_port)
                continue;

            if (!p_port->is_data_worthy())
                continue;

            if (p_port->isSpecialPort())
                continue;

            //get direct route for this port
            direct_route_t *p_curr_direct_route = this->GetDR(p_port);
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s, port=%u",
                                   p_node->getName().c_str(), (u_int8_t)p_port->num);

                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            progress_bar.push(p_port);
            clbck_data.m_data1 = p_port;

            this->ibis_obj.SMPQosConfigVLGetByDirect(p_curr_direct_route,
                                                     &qos_config_vl,
                                                     &clbck_data,
                                                     p_port->num,
                                                     0); // profile_number_enable - ignore in Get()

            if (ibDiagClbck.GetState())
                goto exit;

        } // ports loop
    } // nodes loop

    exit:

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpQoSConfigVLToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_QOS_CONFIG_VL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,"
               "PortGUID,"
               "PortNum";

    for (int i = 0; i < IB_NUM_VL - 1; ++i)
        sstream << ',' << "disable_hoq_life_"  << i
                << ',' << "data_type_bitmask_" << i
                << ',' << "vl_buffer_weight_"  << i;

    sstream << endl;

    csv_out.WriteBuf(sstream.str());

    for (auto node_pair : this->discovered_fabric.NodeByName) {

        IBNode *p_node = node_pair.second;
        if (!p_node)
            continue;

        ib_extended_node_info* p_eni =
                this->fabric_extended_info.getSMPExtNodeInfo(p_node->createIndex);

        if (!p_eni)
            continue;

        // only for Switches and GPUs
        if (p_node->type != IB_SW_NODE && p_eni->node_type_extended != IB_GPU_NODE)
            continue;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {

            IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            struct SMP_QosConfigVL *p_data =
                    this->fabric_extended_info.getSMPQosConfigVL(p_port->createIndex);

            if (!p_data)
                continue;

            sstream.str("");
            sstream << PTR(p_node->guid_get())  << ','
                    << PTR(p_port->guid_get())  << ','
                    << DEC(p_port->num)         ;

            for (int i = 0; i < IB_NUM_VL - 1; ++i)
                sstream << ',' << DEC(p_data->VLConfigBlock[i].disable_hoq_life)
                        << ',' << PTR(p_data->VLConfigBlock[i].data_type_bitmask)
                        << ',' << DEC(p_data->VLConfigBlock[i].vl_buffer_weight);

            sstream << endl;

            csv_out.WriteBuf(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_QOS_CONFIG_VL);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpVPortQoSConfigSLToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_VPORTS_QOS_CONFIG_SL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,PortNum,VPortGUID,VPortNum,SL,BandwidthShare,RateLimit" << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_QosConfigSL *p_curr_vport_qos_config_sl = NULL;
    char buffer[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        if (p_curr_node->type != IB_CA_NODE)
            continue;

        bool vport_rate_limit_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLVPortRateLimitSupported);

        bool vport_bw_alloc_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLVPortAllocBWSupported);

        // go over all ports of this node:
        for (u_int32_t pI = 0; pI <= p_curr_node->numPorts; ++pI) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)pI);

            if (!p_curr_port)
                continue;
            if (!p_curr_port->is_data_worthy())
                continue;

            map_vportnum_vport vports = p_curr_port->VPorts;
            for (map_vportnum_vport::iterator vpI = vports.begin(); vpI != vports.end(); ++vpI) {
                 IBVPort *p_vport = vpI->second;

                 if(!p_vport)
                    continue;

                 p_curr_vport_qos_config_sl =
                              fabric_extended_info.getSMPVPortQosConfigSL(p_vport->createIndex);
                 if (!p_curr_vport_qos_config_sl)           //info not exist
                     continue;

                 for (u_int32_t slI = 0; slI < IB_NUM_SL; ++slI) {
                     sstream.str("");
                     snprintf(buffer, sizeof(buffer),
                              U64H_FMT "," U64H_FMT",%d," U64H_FMT ",%d,%d,",
                              p_curr_node->guid_get(),
                              p_curr_port->guid_get(),
                              p_curr_port->num,
                              p_vport->guid_get(),
                              p_vport->getVPortNum(),
                              slI);

                     sstream << buffer;
                     vport_bw_alloc_flag?
                     sstream << p_curr_vport_qos_config_sl->BandwidthPerSL[slI].BandwidthShare :
                     sstream << "N/A";
                     sstream << ",";
                     vport_rate_limit_flag?
                     sstream << p_curr_vport_qos_config_sl->BandwidthPerSL[slI].RateLimit :
                     sstream << "N/A";
                     sstream << endl;

                     csv_out.WriteBuf(sstream.str());
                 }
            }
        }
    }

    csv_out.DumpEnd(SECTION_VPORTS_QOS_CONFIG_SL);
    IBDIAG_RETURN(rc);
}

int IBDiag::CheckAPortsQosSymmetry(list_p_fabric_general_err& aport_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    for (auto& aports_v_sys_guid_pair : this->GetDiscoverFabricPtr()->APortsBySysGuid) {
        auto& NodeBySystemGuid = this->GetDiscoverFabricPtr()->NodeBySystemGuid;
        auto curr_node_it = NodeBySystemGuid.find(aports_v_sys_guid_pair.first);
        if (curr_node_it == NodeBySystemGuid.end())
            continue;

        IBNode *p_curr_node = curr_node_it->second;
        if (!p_curr_node || p_curr_node->type != IB_CA_NODE)
            continue;

        bool rate_limit_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLRateLimitSupported);

        bool bw_alloc_flag = this->capability_module.IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsQoSConfigSLAllocBWSupported);

        // check that QoS values in all SLs are the same for all planes in a single APort
        for (APort *p_aport : aports_v_sys_guid_pair.second) {
            if (!p_aport || !p_aport->is_data_worthy() || !p_aport->isSymmetric())
                continue;

            if (CheckAPortQosSymmetry(aport_errors, p_aport, rate_limit_flag, bw_alloc_flag))
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
        
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::CheckAPortQosSymmetry(list_p_fabric_general_err& aport_errors,
        const APort *p_aport, bool rate_limit_flag, bool bw_alloc_flag)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (u_int32_t slI = 0; slI < IB_NUM_SL; ++slI) {
        if (rate_limit_flag) {
            // accessor to the rate limit on the curren SL
            ibport_accessor_t<u_int32_t> getRateLimit = std::bind(
                        &IBDMExtendedInfo::getPortQoSRateLimitBySL,
                        &this->fabric_extended_info,
                        std::placeholders::_1, slI);

            if (!p_aport->isEqualAttribute(getRateLimit)) {
                // report error
                string values_arr = p_aport->getAttributeArrayStr<u_int32_t>(
                    getRateLimit, std::to_string);
                aport_errors.push_back(
                    new FabricErrAPortUnequalQoSRateLimit(p_aport, slI, values_arr));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
        if (bw_alloc_flag) {
            // accessor to the bandwidth on the curren SL
            ibport_accessor_t<u_int16_t> getBandwidth = std::bind(
                        &IBDMExtendedInfo::getPortQoSBandwidthBySL,
                        &this->fabric_extended_info,
                        std::placeholders::_1, slI);

            if (!p_aport->isEqualAttribute(getBandwidth)) {
                // report error
                string values_arr = p_aport->getAttributeArrayStr<u_int32_t>(
                    getBandwidth, std::to_string);
                aport_errors.push_back(
                    new FabricErrAPortUnequalQoSBandwidth(p_aport, slI, values_arr));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::BuildSMPTempSensing(
        list_p_fabric_general_err& temp_sensing_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &temp_sensing_errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::IBDiagSMPTempSensingGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect counters
    struct SMP_TempSensing p_temp_sense;
    CLEAR_STRUCT(p_temp_sense);
    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        struct SMP_NodeInfo * p_curr_node_info =
        this->fabric_extended_info.getSMPNodeInfo(
                p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError(
                    "DB error - failed to get SMPNodeInfo for node: %s",
                    p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!this->capability_module.IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsTemperatureSensingSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "node %s: DeviceID %u (0x%x) Does not support"  \
                    " Temprature Sensing, skipping \n",
                    p_curr_node->name.c_str(),
                    p_curr_node_info->DeviceID,
                    p_curr_node_info->DeviceID);
            continue;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);
        //get direct route for this port
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            continue;
        }

        this->ibis_obj.SMPTempSensingDataGetByDirect(
                p_curr_direct_route,
                &p_temp_sense,
                &clbck_data);
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!temp_sensing_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpTempSensingToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_TEMP_SENSING))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID," << "CurrentTemperature" << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_TempSensing *p_curr_temp_sensing = NULL;
    char buffer[1024];

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        p_curr_temp_sensing = this->fabric_extended_info.getSMPTempSensing(i);
        if (!p_curr_temp_sensing)           //info not exist
            continue;

        sstream.str("");
        snprintf(buffer, sizeof(buffer), U64H_FMT ",%d",
                 p_curr_node->guid_get(),
                 (int32_t)p_curr_temp_sensing->current_temp);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_TEMP_SENSING);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildRouterInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect counters
    struct SMP_RouterInfo router_info;
    CLEAR_STRUCT(router_info);

    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_RTR_NODE)
            continue;

        //TODO: add check of cpapbility mask, if the mad is supported

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRouterInfoGetClbck>;

        //get direct route for this port
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            continue;
        }

        progress_bar.push(p_curr_node);
        this->ibis_obj.SMPRouterInfoGetByDirect(
                p_curr_direct_route,
                &router_info,
                &clbck_data);
        if (ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildAdjSubnetsRouterLIDInfoTable(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPAdjRouterLIDInfoTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                            (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_RTR_NODE)
            continue;

        if (!this->capability_module.IsSupportedSMPCapability(p_curr_node,
                                                            EnSMPCapIsRouterLIDSupported))
            continue;

        const SMP_RouterInfo* p_router_info =
                    this->fabric_extended_info.getSMPRouterInfo(p_curr_node->createIndex);
        if (!p_router_info ||
                (!p_router_info->global_router_lid_base &&
                !p_router_info->global_router_lid_top) ||
                !p_router_info->AdjacentSubnetsRouterLIDInfo)
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        static const size_t size = ARRAY_SIZE(SMP_AdjSubnetsRouterLIDInfoTable::Record);
        uint8_t num_of_blocks = static_cast<uint8_t>(
                NUM_OF_ROUTER_TABLE_ENTRIES(p_router_info->AdjacentSubnetsRouterLIDInfo, size));
        for (u_int8_t i = 0; i < num_of_blocks; i++) {
            SMP_AdjSubnetsRouterLIDInfoTable  adj_router_lid_table;
            clbck_data.m_data2 = (void *)(uintptr_t)i;
            this->ibis_obj.SMPAdjSubnetRouterLIDInfoTableGetByDirect(p_curr_direct_route,
                                                    i, //send block num
                                                    &adj_router_lid_table,
                                                    &clbck_data);
            if (ibDiagClbck.GetState())
                break;
        }//loop over adjust subnets
    }//loop over routers

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildARGroupToRouterLIDTable(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPARGroupToRouterLidTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (const IBNode *p_router : this->discovered_fabric.Routers) {
        if (!p_router) {
            this->SetLastError("DB error - found null node in Routers set");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_router->getInSubFabric() ||
                !this->capability_module.IsSupportedSMPCapability(p_router,
                                                                  EnSMPCapIsRouterLIDSupported))
            continue;

        const SMP_RouterInfo* p_router_info =
                    this->fabric_extended_info.getSMPRouterInfo(p_router->createIndex);
        if (!p_router_info ||
                (!p_router_info->global_router_lid_base &&
                !p_router_info->global_router_lid_top) ||
                !p_router_info->pfrn_rtr_en ||
                !p_router_info->max_ar_group_id)
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_router);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_router->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = (void*)(p_router);
        progress_bar.push(p_router);

        const static size_t size = ARRAY_SIZE(SMP_ARGroupToRouterLIDTable::router_lid);
        uint8_t num_of_blocks = static_cast<uint8_t>(
                NUM_OF_ROUTER_TABLE_ENTRIES(p_router_info->max_ar_group_id, size));
        for (u_int8_t i = 0; i < num_of_blocks; i++) {
            SMP_ARGroupToRouterLIDTable record;
            clbck_data.m_data2 = (void *)(uintptr_t)i;
            this->ibis_obj.SMPARGroupToRouterLIDTableGetByDirect(p_curr_direct_route,
                                                                 i, //send block num
                                                                 &record,
                                                                 &clbck_data);
            if (ibDiagClbck.GetState())
                break;
        }
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildRouterLIDTable(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPRouterLIDTableGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                            (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (p_curr_node->type != IB_RTR_NODE || !p_curr_node->getInSubFabric())
            continue;

        if (!this->capability_module.IsSupportedSMPCapability(p_curr_node,
                                                            EnSMPCapIsRouterLIDSupported))
            continue;

        const SMP_RouterInfo* p_router_info =
                    this->fabric_extended_info.getSMPRouterInfo(p_curr_node->createIndex);
        if (!p_router_info ||
                (!p_router_info->global_router_lid_base &&
                !p_router_info->global_router_lid_top))
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        const size_t size = ARRAY_SIZE(SMP_RouterLIDTable::V);
        uint8_t start_block = static_cast<uint8_t>(p_router_info->global_router_lid_base / size);
        uint8_t end_block = static_cast<uint8_t>(p_router_info->global_router_lid_top / size);
        for (uint8_t i = start_block; i <= end_block; i++) {
            SMP_RouterLIDTable  router_lid_table;
            clbck_data.m_data2 = (void *)(uintptr_t)i;
            this->ibis_obj.SMPRouterLIDTableGetByDirect(p_curr_direct_route,
                                            i, //send block num
                                            &router_lid_table,
                                            &clbck_data);
            if (ibDiagClbck.GetState())
                break;
        }
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpAdjSubnetsRouterLIDInfoTableToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    if (csv_out.DumpStart(SECTION_ROUTERS_ADJ_LID_TBL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,BlockNum,RecordNum,subnet_prefix_id,local_router_LID_start,"
            << "local_router_LID_start_cont,local_router_LID_end"
            << endl;
    csv_out.WriteBuf(sstream.str());

    //go over all Routers
    for (set_pnode::iterator nI = this->discovered_fabric.Routers.begin();
            nI != this->discovered_fabric.Routers.end(); ++nI) {
        const IBNode *p_router = *nI;
        if (!p_router) {
            this->SetLastError("DB error - found null node in Routers set");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_router->getInSubFabric())
            continue;

        const SMP_RouterInfo *p_router_info = this->fabric_extended_info.getSMPRouterInfo(p_router->createIndex);
        if (!p_router_info)
            continue;

        const size_t size = ARRAY_SIZE(SMP_AdjSubnetsRouterLIDInfoTable::Record);
        u_int8_t block_num = 0;
        u_int8_t idx = 0;
        const SMP_AdjSubnetsRouterLIDInfoTable *p_adj_rtr_tbl = NULL;
        for (u_int8_t j = 0; j < p_router_info->AdjacentSubnetsRouterLIDInfo; j++) {
            // get the next block, initialized already to first
            if (j % size == 0) {
                block_num = static_cast<u_int8_t>(j / size);
                p_adj_rtr_tbl =
                    this->fabric_extended_info.getSMPAdjSubnteRouterLIDInfoTbl(p_router->createIndex,
                                                                               block_num);
            }

            if (!p_adj_rtr_tbl)
                continue;

            idx = static_cast<u_int8_t>(j % size);
            sstream.str("");
            sstream << PTR(p_router->guid_get()) << ','
                    << +block_num                << ','
                    << +idx                      << ','
                    << PTR(p_adj_rtr_tbl->Record[idx].subnet_prefix_id)        << ','
                    << p_adj_rtr_tbl->Record[idx].local_router_lid_start       << ','
                    << +p_adj_rtr_tbl->Record[idx].local_router_lid_start_cont << ','
                    << p_adj_rtr_tbl->Record[idx].local_router_lid_end
                    << endl;
            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_ROUTERS_ADJ_LID_TBL);

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteARGroupToRouterFLIDData(const string& file_name)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream  outStream;
    int rc = this->OpenFile("AR Groups to FLIDs", file_name, outStream, false, true);
    if (rc)
        IBDIAG_RETURN(rc);

    if (!outStream.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    outStream << "#Adaptive Routing Groups To FLIDs" << endl;

    for (const IBNode *p_router : this->discovered_fabric.Routers) {
        if (!p_router) {
            this->SetLastError("DB error - found null node in Routers set");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        const SMP_RouterInfo* p_router_info =
                    this->fabric_extended_info.getSMPRouterInfo(p_router->createIndex);
        if (!p_router_info)
            continue;

        outStream << endl
                  << "Router: " << PTR(p_router->guid_get())
                  << " - " << p_router->getName() << endl
                  << "AR Group         "  << "FLID" << endl;

        const static size_t size = ARRAY_SIZE(SMP_ARGroupToRouterLIDTable::router_lid);
        u_int8_t block_num = 0;
        u_int8_t idx = 0;
        const SMP_ARGroupToRouterLIDTable *p_record = NULL;
        for (u_int16_t j = 0; j < p_router_info->max_ar_group_id; j++) {
            // get the next block, initialized already to first
            if (j % size == 0) {
                block_num = static_cast<u_int8_t>(j / size);
                p_record =
                    this->fabric_extended_info.getSMPARGroupToRouterLIDTbl(p_router->createIndex,
                                                                           block_num);
            }

            if (!p_record)
                continue;

            idx = static_cast<u_int8_t>(j % size);
            //Router_lid[0] will reflect the FLID AR of group #[block_index * 32]
            //Router_lid[31] will reflect the FLID of AR group #([block_index * 32] + 31)
            //i.e. #group = (block_num * size) + idx;
            if (!p_record->router_lid[idx])
                continue; //may be we can break here...

            outStream << (block_num * size) + idx
                    << "                   "
                    << p_record->router_lid[idx]
                    << endl;
        }
    }

    this->CloseFile(outStream);

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildAdjSiteLocalSubnetsTable(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes and collect counters
    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric() || p_curr_node->type != IB_RTR_NODE)
            continue;

        struct SMP_RouterInfo *p_router_info = this->fabric_extended_info.
                                               getSMPRouterInfo(
                                                   p_curr_node->createIndex);
        // the info doesn't exist or there are no entries
        // in neither adjacent site local subnets router table
        // nor in the next hop table
        if (!p_router_info ||
            (p_router_info->AdjacentSiteLocalSubnetsTableTop == 0 &&
             p_router_info->NextHopTableTop == 0))
            continue;

        u_int8_t adj_num_blocks =
            (u_int8_t)NUM_OF_ROUTER_TABLE_ENTRIES(
                          p_router_info->AdjacentSiteLocalSubnetsTableTop,
                          IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK);
        //TODO: add check of cpapbility mask, if the mad is supported

        clbck_data.m_data1 = p_curr_node;

        //get direct route for this port
        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPAdjRouterTableGetClbck>;

        struct SMP_AdjSiteLocalSubnTbl adj_router_table;
        CLEAR_STRUCT(adj_router_table);
        for (u_int8_t i = 0; i < adj_num_blocks; i++) {
            //save the block number
            clbck_data.m_data2 = (void *)(uintptr_t)i;
            progress_bar.push(p_curr_node);
            this->ibis_obj.SMPAdjRouterTableGetByDirect(
                        p_curr_direct_route,
                        i, //send block num
                        &adj_router_table,
                        &clbck_data);
            if (ibDiagClbck.GetState())
                goto exit;
        }

        // now get the next hop table
        u_int32_t nh_num_blocks = (u_int32_t)NUM_OF_ROUTER_TABLE_ENTRIES(
                                      p_router_info->NextHopTableTop,
                                      IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK);

        clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPNextHopRouterTableGetClbck>;

        struct SMP_NextHopTbl nh_router_table;
        CLEAR_STRUCT(nh_router_table);
        for (u_int32_t j = 0; j < nh_num_blocks; j++) {
            //save the block number
            clbck_data.m_data2 = (void *)(uintptr_t)j;
            progress_bar.push(p_curr_node);
            this->ibis_obj.SMPNextHopRouterTableGetByDirect(
                        p_curr_direct_route,
                        j, //send block num
                        &nh_router_table,
                        &clbck_data);
            if (ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpRouterInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_ROUTERS_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,CapabilityMask,NextHopTableCap,NextHopTableTop,"
            << "AdjacentSiteLocalSubnetsTableTop,AdjacentSiteLocalSubnetsTableCap,"
            << "table_changes_bitmask,ar_group_to_router_lid_table_cap,"
            << "cap_supported_subnets,cap_router_lid,"
            << "AdjacentSubnetsRouterLIDInfo,pfrn_rtr_en,"
            << "global_router_lid_base,max_ar_group_id,global_router_lid_top,"
            << "local_router_lid_base,local_router_lid_top"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); i++) {
        const IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        const SMP_RouterInfo *p_router_info = this->fabric_extended_info.getSMPRouterInfo(i);
        if (!p_router_info)           //info not exist
            continue;

        sstream.str("");
        sstream << PTR(p_curr_node->guid_get())                   <<  ','
                << p_router_info->CapabilityMask                  <<  ','
                << p_router_info->NextHopTableCap                 <<  ','
                << p_router_info->NextHopTableTop                 <<  ','
                << +p_router_info->AdjacentSiteLocalSubnetsTableTop  <<  ','
                << +p_router_info->AdjacentSiteLocalSubnetsTableCap  <<  ','
                << p_router_info->table_changes_bitmask           <<  ','
                << p_router_info->ar_group_to_router_lid_table_cap   <<  ','
                << +p_router_info->cap_supported_subnets          <<  ','
                << p_router_info->cap_router_lid                  <<  ','
                << +p_router_info->AdjacentSubnetsRouterLIDInfo   <<  ','
                << +p_router_info->pfrn_rtr_en                    <<  ','
                << p_router_info->global_router_lid_base          <<  ','
                << p_router_info->max_ar_group_id                 <<  ','
                << p_router_info->global_router_lid_top           <<  ','
                << p_router_info->local_router_lid_base           <<  ','
                << p_router_info->local_router_lid_top            <<  endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_ROUTERS_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpRouterAdjSiteLocalSubnetTableToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_ROUTERS_ADJ_TBL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,BlockNum,RecordNum,SubnetPrefix,Pkey,MasterSMLID,"
            << "max_ar_group_id,pfrn_rtr_en" << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_RouterInfo *p_router_info = NULL;
    char buffer[1024] = {'\0'};

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize();
         i++) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        p_router_info = this->fabric_extended_info.getSMPRouterInfo(i);
        if (!p_router_info)           //info not exist
            continue;

        u_int8_t num_entries =
            p_router_info->AdjacentSiteLocalSubnetsTableTop;

        u_int8_t block_num = 0, idx = 0;
        SMP_AdjSiteLocalSubnTbl *p_adj_rtr_tbl = NULL;
        for (u_int8_t j = 0; j < num_entries; j++) {
                /* get the next block, initialized already to first */
                if ((j % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK) == 0) {
                        block_num = j / IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK;
                        p_adj_rtr_tbl = this->fabric_extended_info.
                                                  getSMPAdjSiteLocalSubnTbl(
                                                      i, block_num);
                }
                if (!p_adj_rtr_tbl)
                    continue;

                idx = (u_int8_t)(j % IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK);
                sstream.str("");
                snprintf(buffer, sizeof(buffer),
                         U64H_FMT "," U8H_FMT "," U8H_FMT "," U16H_FMT ","
                         U16H_FMT "," U16H_FMT "," U16H_FMT "," U8H_FMT,
                         p_curr_node->guid_get(),
                         block_num, idx, p_adj_rtr_tbl->Record[idx].SubnetPrefix,
                         p_adj_rtr_tbl->Record[idx].Pkey,
                         p_adj_rtr_tbl->Record[idx].MasterSMLID,
                         p_adj_rtr_tbl->Record[idx].max_ar_group_id,
                         p_adj_rtr_tbl->Record[idx].pfrn_rtr_en);
                sstream << buffer << endl;
                csv_out.WriteBuf(sstream.str());
        }

    }

    csv_out.DumpEnd(SECTION_ROUTERS_ADJ_TBL);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpRouterNextHopToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_ROUTERS_NEXT_HOP_TBL))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,BlockNum,RecordNum,SubnetPrefix,Pkey,Weight" << endl;
    csv_out.WriteBuf(sstream.str());

    struct SMP_RouterInfo *p_router_info = NULL;
    char buffer[1024] = {'\0'};

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize();
         i++) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;
        p_router_info = this->fabric_extended_info.getSMPRouterInfo(i);
        if (!p_router_info)           //info not exist
            continue;

        u_int32_t num_entries =
            p_router_info->NextHopTableTop;

        u_int32_t block_num = 0, idx = 0;
        SMP_NextHopTbl *p_nh_rtr_tbl = NULL;
        for (u_int32_t j = 0; j < num_entries; j++) {
                /* get the next block, initialized already to first */
                if ((j % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK) == 0) {
                        block_num = j / IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK;
                        p_nh_rtr_tbl = this->fabric_extended_info.
                                                  getSMPNextHopTbl(
                                                      i, block_num);
                }
                if (!p_nh_rtr_tbl)
                    continue;

                idx = (u_int32_t)(j % IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK);
                sstream.str("");
                snprintf(buffer, sizeof(buffer),
                         U64H_FMT "," U32H_FMT "," U32H_FMT "," U64H_FMT ","
                         U16H_FMT "," U8H_FMT,
                         p_curr_node->guid_get(),
                         block_num, idx, p_nh_rtr_tbl->Record[idx].SubnetPrefix,
                         p_nh_rtr_tbl->Record[idx].Pkey,
                         p_nh_rtr_tbl->Record[idx].Weight);
                sstream << buffer << endl;
                csv_out.WriteBuf(sstream.str());
        }

    }
    csv_out.DumpEnd(SECTION_ROUTERS_NEXT_HOP_TBL);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


/*** Fast Recovery Functions ***/

// Profiles Config Functions
int IBDiag::BuildProfilesConfig(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;
    clbck_data_t clbck_data;

    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMP_ProfilesConfigGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_ProfilesConfig profiles_config;
    CLEAR_STRUCT(profiles_config);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsFastRecoverySupported))
            continue;

        direct_route_t *p_direct_route = this->GetDR(p_curr_node);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        int num_blocks = p_curr_node->numPorts / PORT_PROFILES_IN_BLOCK + 1;

        for (int block = 0; block < num_blocks ; block++) {

            clbck_data.m_data2 = (void *)(uintptr_t)block;

            this->ibis_obj.SMPProfilesConfigGetByDirect(p_direct_route,
                                                        block,
                                                        FAST_RECOVERY_PROFILE,
                                                        &profiles_config,
                                                        &clbck_data);
            if (ibDiagClbck.GetState())
                break;

        } // blocks
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpProfilesConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_PROFILES_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    SMP_ProfilesConfig* p_profiles_config;

    stringstream sstream;
    sstream << "NodeGUID,PortNumber,PortProfile" << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        int num_blocks = p_curr_node->numPorts / PORT_PROFILES_IN_BLOCK + 1;

        for (int block = 0; block < num_blocks ; block++) {

            p_profiles_config = this->fabric_extended_info.
                getProfilesConfig(p_curr_node->createIndex, block);
            if (!p_profiles_config) //info not exist
                continue;

            for (int port = 0;
                    port < PORT_PROFILES_IN_BLOCK &&
                        (block * PORT_PROFILES_IN_BLOCK + port) <= p_curr_node->numPorts;
                    ++port) {

                IBPort* p_port = p_curr_node->getPort((phys_port_t)(block * PORT_PROFILES_IN_BLOCK + port));
                if (p_port && p_port->isSpecialPort())
                    continue;

                sstream.str("");

                sstream << PTR(p_curr_node->guid_get())            << ","
                        << block * PORT_PROFILES_IN_BLOCK + port   << ","
                        << +p_profiles_config->port_profile[port]  << endl;

                csv_out.WriteBuf(sstream.str());
            } // ports
        } // blocks
    } // switches

    csv_out.DumpEnd(SECTION_PROFILES_CONFIG);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Credit Watchdog Config Functions
int IBDiag::BuildCreditWatchdogConfig(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;
    clbck_data_t clbck_data;

    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMP_CreditWatchdogConfigGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_CreditWatchdogConfig credit_wd_config;
    CLEAR_STRUCT(credit_wd_config);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsCreditWatchdogSupported))
            continue;

        direct_route_t *p_direct_route = this->GetDR(p_curr_node);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        for (auto profile : p_curr_node->fast_recovery_profiles) {

            clbck_data.m_data2 = (void *)(uintptr_t)profile;

            this->ibis_obj.SMPCreditWatchdogConfigGetByDirect(p_direct_route,
                                                              profile,
                                                              &credit_wd_config,
                                                              &clbck_data);
            if (ibDiagClbck.GetState())
                break;

        } // profiles
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpCreditWatchdogConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_CREDIT_WATCHDOG_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    SMP_CreditWatchdogConfig* p_credit_wd_config;

    stringstream sstream;
    sstream << "NodeGUID,ProfileNum,en_thr,error_thr_action,"
            << "en_normal_trap,en_warning_trap,en_error_trap,"
            << "error_thr,warning_thr,normal_thr,"
            << "time_window,sampling_rate"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (auto profile : p_curr_node->fast_recovery_profiles) {

            p_credit_wd_config = this->fabric_extended_info.
                getCreditWatchdogConfig(p_curr_node->createIndex, profile);
            if (!p_credit_wd_config) // info not exist
                continue;

            sstream.str("");

            sstream << PTR(p_curr_node->guid_get())           << ","
                    << +profile                               << ","
                    << +p_credit_wd_config->en_thr            << ","
                    << +p_credit_wd_config->error_thr_action  << ","
                    << +p_credit_wd_config->en_normal_trap    << ","
                    << +p_credit_wd_config->en_warning_trap   << ","
                    << +p_credit_wd_config->en_error_trap     << ","
                    << +p_credit_wd_config->error_thr         << ","
                    << +p_credit_wd_config->warning_thr       << ","
                    << +p_credit_wd_config->normal_thr        << ","
                    << +p_credit_wd_config->time_window       << ","
                    << +p_credit_wd_config->sampling_rate     << endl;

            csv_out.WriteBuf(sstream.str());
        } // profiles
    } // switches

    csv_out.DumpEnd(SECTION_CREDIT_WATCHDOG_CONFIG);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// BER Config Functions
int IBDiag::BuildBERConfig(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;
    clbck_data_t clbck_data;

    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = &forwardClbck<IBDiagClbck, &IBDiagClbck::SMP_BERConfigGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_BERConfig ber_config;
    CLEAR_STRUCT(ber_config);

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsBERConfigSupported))
            continue;

        direct_route_t *p_direct_route = this->GetDR(p_curr_node);
        if (!p_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());

            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        for (auto profile : p_curr_node->fast_recovery_profiles) {

            clbck_data.m_data2 = (void *)(uintptr_t)profile;

            for (int ber_type = 0; ber_type < NUM_OF_BER_TYPES ; ber_type++) {

                clbck_data.m_data3 = (void *)(uintptr_t)ber_type;

                this->ibis_obj.SMPBERConfigGetByDirect(p_direct_route,
                                                       profile,
                                                       ber_type,
                                                       0, //default_thr
                                                       &ber_config,
                                                       &clbck_data);
                if (ibDiagClbck.GetState())
                    break;

            } // BER types
        } // profiles
    } // switches

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpBERConfigToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_BER_CONFIG))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    SMP_BERConfig* p_ber_config;

    stringstream sstream;
    sstream << "NodeGUID,ProfileNum,BERType,en_thr,error_thr_action,"
            << "en_normal_trap,en_warning_trap,en_error_trap,"
            << "error_thr,warning_thr,normal_thr,"
            << "time_window,sampling_rate"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
            nI != this->discovered_fabric.Switches.end(); ++nI) {

        IBNode *p_curr_node = *nI;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (auto profile : p_curr_node->fast_recovery_profiles) {
            for (int ber_type = 0; ber_type < NUM_OF_BER_TYPES ; ber_type++) {

                p_ber_config = this->fabric_extended_info.getBERConfig(p_curr_node->createIndex,
                                                                       profile,
                                                                       ber_type);
                if (!p_ber_config) // info not exist
                    continue;

                sstream.str("");

                sstream << PTR(p_curr_node->guid_get())      << ","
                        << +profile                          << ","
                        << ber_type                          << ","
                        << +p_ber_config->en_thr             << ","
                        << +p_ber_config->error_thr_action   << ","
                        << +p_ber_config->en_normal_trap     << ","
                        << +p_ber_config->en_warning_trap    << ","
                        << +p_ber_config->en_error_trap      << ","
                        << +p_ber_config->error_thr          << ","
                        << +p_ber_config->warning_thr        << ","
                        << +p_ber_config->normal_thr         << ","
                        << +p_ber_config->time_window        << ","
                        << +p_ber_config->sampling_rate      << endl;

                csv_out.WriteBuf(sstream.str());

            } // BER types
        } // profiles
    } // switches

    csv_out.DumpEnd(SECTION_BER_CONFIG);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildExtendedSwitchInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMP_ExtendedSwitchInfoClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_ExtendedSwitchInfo smp_extended_switch_info;
    CLEAR_STRUCT(smp_extended_switch_info);

    for (IBNode *p_curr_node : this->discovered_fabric.Switches) {
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsExtendedSwitchIfoSupported))
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.SMPExtendedSwitchInfoMadGetSetByDirect(p_curr_direct_route, 
                                                              IBIS_IB_MAD_METHOD_GET, 
                                                              &smp_extended_switch_info, 
                                                              &clbck_data);
    }
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpExtendedSwitchInfoToCSV(CSVOut &csv_out) {
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_EXTENDED_SWITCH_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,end_to_end_timescale,request_issu,turbo_path_cap,turbo_path_enable,"
            << "req_delay_cap,set_trig_th_cap,rst_trig_th_cap,req_trig_window_cap,"
            << "req_delay,set_trig_th,rst_trig_th,req_trig_window" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)
            continue;

        SMP_ExtendedSwitchInfo* extended_switch_info = this->fabric_extended_info.
                                    getSMP_ExtendedSwitchInfo(p_curr_node->createIndex);
        if (!extended_switch_info)
            continue;

        sstream.str("");
        sstream << PTR(p_curr_node->guid_get())                << ','
                << +extended_switch_info->end_to_end_timescale << ','
                << +extended_switch_info->request_issu         << ','
                << +extended_switch_info->turbo_path_cap       << ','
                << +extended_switch_info->turbo_path_enable    << ','
                << +extended_switch_info->req_delay_cap        << ','
                << +extended_switch_info->set_trig_th_cap      << ','
                << +extended_switch_info->rst_trig_th_cap      << ','
                << +extended_switch_info->req_trig_window_cap  << ','
                << +extended_switch_info->req_delay            << ','
                << +extended_switch_info->set_trig_th          << ','
                << +extended_switch_info->rst_trig_th          << ','
                << +extended_switch_info->req_trig_window      << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_EXTENDED_SWITCH_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::BuildChassisInfo(list_p_fabric_general_err& errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &errors);

    ProgressBarNodes progress_bar;

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPChassisInfoClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    struct SMP_ChassisInfo smp_chassis_info = {};

    for (auto node_pair : this->discovered_fabric.NodeByName) {

        IBNode *p_curr_node = node_pair.second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!this->GetCapabilityModulePtr()->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsChassisInfoSupported))
            continue;

        direct_route_t *p_curr_direct_route = this->GetDR(p_curr_node);
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;
        progress_bar.push(p_curr_node);

        this->ibis_obj.SMPChassisInfoMadGetByDirect(p_curr_direct_route,
                                                    &smp_chassis_info,
                                                    &clbck_data);
    } // nodes

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpChassisInfoToCSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (csv_out.DumpStart(SECTION_CHASSIS_INFO))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    stringstream sstream;
    sstream << "NodeGUID,chassis_id"
            << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {

        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)
            continue;

        SMP_ChassisInfo* p_chassis_info = this->fabric_extended_info.
                                    getSMPChassisInfo(p_curr_node->createIndex);
        if (!p_chassis_info)
            continue;

        sstream.str("");
        sstream << PTR(p_curr_node->guid_get())  << ','
                << p_chassis_info->chassis_id    << endl;

        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CHASSIS_INFO);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
