 /*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <math.h>
#include "diagnostic_data.h"
#include "acc_reg.h"
#include "infiniband/ibutils/output_stream_primitives.h"

#define HILOW_TO_UINT64(x) ((uint64_t(x##_high) << 32) + x##_low)

DiagnosticDataInfo::DiagnosticDataInfo(int page_id, int support_version,
                               int num_fields, const string &name,
                               u_int64_t not_supported_bit, u_int32_t dd_type,
                               const string& header, u_int8_t is_per_node,
                               PhyPluginSupportedNodesType support_nodes,
                               bool enable_disconnected_ports, 
                               const string& common_header):
                               m_page_id(page_id),
                               m_support_version(support_version),
                               m_num_fields(num_fields),
                               m_not_supported_bit(not_supported_bit),
                               m_dd_type(dd_type),
                               m_header(header),
                               m_is_per_node(is_per_node),
                               m_support_nodes(support_nodes),
                               m_p_ibdiag(NULL),
                               m_name(name),
                               m_common_header(common_header),
                               m_enable_disconnected_ports(enable_disconnected_ports)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    int status = csv_out.DumpStart(this->GetSectionHeader().c_str());
    if (status)
        IBDIAG_RETURN(status);

    sstream << m_common_header;

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataInfo::DumpDiagnosticDataHeaderEnd(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    csv_out.DumpEnd(this->GetSectionHeader().c_str());

    IBDIAGNET_RETURN_VOID;
}

bool DiagnosticDataInfo::IsDDPageSupportedInNode(struct DDPageIdentification *p_page_identification)
{
    IBDIAGNET_ENTER;

    u_int32_t group = this->m_page_id / (sizeof(p_page_identification->pages_group[0]) * 8);
    u_int32_t bit =  this->m_page_id % (sizeof(p_page_identification->pages_group[0]) * 8);

    IBDIAGNET_RETURN(p_page_identification->pages_group[group] & (1 << bit));
}


/***********************************************
 ***   Diagnostic Data Page Identification   ***
 ***********************************************/
DiagnosticDataPageIdentification::DiagnosticDataPageIdentification():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_PAGE,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_VERSION,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PAGE_IDENTIFICATION,
                       true,
                       SUPPORT_SW,
                       false,
                       "NodeGuid,Version")
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int  DiagnosticDataPageIdentification::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
     IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header;

    for (size_t i = 0; i < ARRAY_SIZE(DDPageIdentification::pages_group); i++)
         sstream << ',' << "PageIdentification.pages_group[" << i << ']';

    sstream << endl;

    csv_out.WriteBuf(sstream.str());
#endif

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPageIdentification::DumpDiagnosticData(stringstream &sstream,
                                                          VS_DiagnosticData &dd,
                                                          IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    //unpack Diagnostic Data page identification
    struct DDPageIdentification page_identification;
    DDPageIdentification_unpack(&page_identification, (uint8_t *)&(dd.data_set));

    for (size_t i = 0;
         i < sizeof(page_identification.pages_group) / sizeof(page_identification.pages_group[0]);
         i++)
            sstream << ',' << PTR(page_identification.pages_group[i]);

    sstream << endl;
    IBDIAGNET_RETURN_VOID
}


/*****************************************
 ***   Diagnostic Data PCI functions   ***
 *****************************************/
DiagnosticDataPCI::DiagnosticDataPCI(int page_id, int support_version,
                                     int num_fields, const string &name, u_int64_t not_supported_bit,
                                     u_int32_t dd_type, string header, u_int8_t is_per_node,
                                     PhyPluginSupportedNodesType support_nodes) :
                                             DiagnosticDataInfo(page_id,
                                                                support_version,
                                                                num_fields,
                                                                name,
                                                                not_supported_bit,
                                                                dd_type,
                                                                header,
                                                                is_per_node,
                                                                support_nodes, false, 
                                                                "NodeGuid,PCIIndex,Depth,PCINode,Version")
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

// Diagnostic Data PCIE Counters
DiagnosticDataPCIECntrs::DiagnosticDataPCIECntrs():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_CNTRS_PAGE,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_VERSION,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_CNTRS_NAME,
                          NSB::get(this),
                          DD_PCI_TYPE,
                          SECTION_PCI_CNTRS,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPCIECntrs::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const {
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);
#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    MPCNT_PCIe_Counters_Register::DumpFieldNames(sstream);
    sstream << endl;

    csv_out.WriteBuf(sstream.str());
#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPCIECntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pcie_perf_counters p_pci_cntrs;
    pcie_perf_counters_unpack(&p_pci_cntrs, (uint8_t *)&(dd.data_set));

    MPCNT_PCIe_Counters_Register::DumpData(p_pci_cntrs, sstream);

    IBDIAGNET_RETURN_VOID;
}

// Diagnostic Data PCIe Timers and States
DiagnosticDataPCIETimers::DiagnosticDataPCIETimers():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_TIMERS_PAGE,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_VERSION,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_TIMERS_NAME,
                          NSB::get(this),
                          DD_PCI_TYPE,
                          SECTION_PCI_TIMERS,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPCIETimers::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const {
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    MPCNT_PCIe_Timers_Register::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPCIETimers::DumpDiagnosticData(stringstream &sstream,
                                                  VS_DiagnosticData &dd,
                                                  IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pcie_timers_states p_pci_timers;
    pcie_timers_states_unpack(&p_pci_timers, (uint8_t *)&(dd.data_set));

    MPCNT_PCIe_Timers_Register::DumpData(p_pci_timers, sstream);

    IBDIAGNET_RETURN_VOID;
}

// Diagnostic Data PCIe Lanes Counters
DiagnosticDataPCIELanes::DiagnosticDataPCIELanes():
        DiagnosticDataPCI(DIAGNOSTIC_DATA_PCIE_LANES_PAGE,
                          DIAGNOSTIC_DATA_PCIE_LANES_VERSION,
                          DIAGNOSTIC_DATA_PCIE_LANES_NUM_FIELDS,
                          DIAGNOSTIC_DATA_PCIE_LANES_NAME,
                          NSB::get(this),
                          DD_PCI_TYPE,
                          SECTION_PCI_LANES,
                          true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPCIELanes::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const {
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    MPCNT_PCIe_Lane_Counters_Register::DumpFieldNames(sstream);
    sstream << endl;

    csv_out.WriteBuf(sstream.str());

#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPCIELanes::DumpDiagnosticData(stringstream &sstream,
                                                 VS_DiagnosticData &dd,
                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pcie_lanes_counters p_pci_lanes;
    pcie_lanes_counters_unpack(&p_pci_lanes, (uint8_t *)&(dd.data_set));

    MPCNT_PCIe_Lane_Counters_Register::DumpData(p_pci_lanes, sstream);

    IBDIAGNET_RETURN_VOID;
}


/****************************************
 ***  Diagnostic Data PHY functions   ***
 ****************************************/
DiagnosticDataPhysLayerCntrs::DiagnosticDataPhysLayerCntrs():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_PAGE,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_VERSION,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PHY_CNTRS,
                       false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiagnosticDataPhysLayerCntrs::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream <<'#' << m_common_header <<  ',';
    PPCNT_Phys_Layer_Cntrs::DumpFieldNames(sstream);
    sstream << endl;

    csv_out.WriteBuf(sstream.str());
#endif

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPhysLayerCntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                      IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    //unpack physical layer counters page
    struct phys_layer_cntrs phy_cntrs;
    phys_layer_cntrs_unpack(&phy_cntrs, (uint8_t *)&(dd.data_set));

    PPCNT_Phys_Layer_Cntrs::DumpData(phy_cntrs, sstream);

    IBDIAGNET_RETURN_VOID;
}

// PDDR Troubleshooting Info
DiagnosticDataTroubleshootingInfo::DiagnosticDataTroubleshootingInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_PAGE,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_VERSION,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_TROUBLESHOOTING_INFO,
                       false)
{
}

int DiagnosticDataTroubleshootingInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const {
   IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

    stringstream sstream;
    PDDRTroubleshootingInfoRegister::DumpRealHeader(sstream, m_common_header);
    sstream << std::endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataTroubleshootingInfo::DumpDiagnosticData(stringstream &sstream,
                                        VS_DiagnosticData &dd, IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pddr_troubleshooting_page troubleshooting_info;
    pddr_troubleshooting_page_unpack(&troubleshooting_info, (uint8_t *)&(dd.data_set));
    PDDRTroubleshootingInfoRegister::DumpLayout(sstream, troubleshooting_info);

    IBDIAGNET_RETURN_VOID;
}

// PDDR Operation Info
DiagnosticDataOperationInfo::DiagnosticDataOperationInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_OPERATION_INFO_PAGE,
                       DIAGNOSTIC_DATA_OPERATION_INFO_VERSION,
                       DIAGNOSTIC_DATA_OPERATION_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_OPERATION_INFO_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_OPERATION_INFO,
                       false)
{
}

int DiagnosticDataOperationInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

    stringstream sstream;
    PDDROperationInfoRegister::DumpRealHeader(sstream, m_common_header);
    sstream << std::endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataOperationInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                     IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pddr_operation_info_page operation_info;
    pddr_operation_info_page_unpack(&operation_info, (uint8_t *)&(dd.data_set));
    PDDROperationInfoRegister::DumpLayout(sstream, operation_info);

    IBDIAGNET_RETURN_VOID;
}

// PDDR Phy Info
DiagnosticDataPhyInfo::DiagnosticDataPhyInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_INFO_PAGE,
                       DIAGNOSTIC_DATA_PHY_INFO_VERSION,
                       DIAGNOSTIC_DATA_PHY_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHY_INFO_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PHY_INFO,
                       false)
{
}

int DiagnosticDataPhyInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

    stringstream sstream;
    PDDRPhyInfoRegister::DumpPhyInfoRealHeader(sstream, m_common_header);
    sstream << std::endl;
    csv_out.WriteBuf(sstream.str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPhyInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                               IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct pddr_phy_info_page phy_info;
    pddr_phy_info_page_unpack(&phy_info, (uint8_t *)&(dd.data_set));

    PDDRPhyInfoRegister::DumpLayout(sstream, phy_info);
    IBDIAGNET_RETURN_VOID;
}

// PDDR Link Down Info
DiagnosticDataLinkDownInfo::DiagnosticDataLinkDownInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LINK_DOWN_INFO_PAGE,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_VERSION,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_LINK_DOWN_INFO,
                       false)
{
}

int DiagnosticDataLinkDownInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

    stringstream sstream;
    PDDRLinkDownInfoRegister::DumpRealHeader(sstream, m_common_header);
    sstream << std::endl;
    csv_out.WriteBuf(sstream.str());

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataLinkDownInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct pddr_link_down_info_page link_down_info;
    pddr_link_down_info_page_unpack(&link_down_info, (uint8_t *)&(dd.data_set));
    PDDRLinkDownInfoRegister::DumpLayout(sstream, link_down_info);

    IBDIAGNET_RETURN_VOID;
}

// PDDR Link Down Info
DiagnosticDataLinkUpInfo::DiagnosticDataLinkUpInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LINK_UP_INFO_PAGE,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_VERSION,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_NUM_FIELDS,
                       DIAGNOSTIC_DATA_LINK_UP_INFO_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_LINK_UP_INFO,
                       false)
{
}

int DiagnosticDataLinkUpInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

    stringstream sstream;
    PDDRLinkUpInfoRegister::DumpLinkUpInfoRealHeader(sstream, m_common_header);
    sstream << std::endl;
    csv_out.WriteBuf(sstream.str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataLinkUpInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct pddr_link_up_info_page link_up_info;
    pddr_link_up_info_page_unpack(&link_up_info, (uint8_t *)&(dd.data_set));
    PDDRLinkUpInfoRegister::DumpLayout(sstream, link_up_info);
    IBDIAGNET_RETURN_VOID;
}


// PPCNT pages
// PPCNT RS FEC Histograms
DiagnosticDataRSHistograms::DiagnosticDataRSHistograms():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_PAGE,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_VERSION,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_RS_FEC_HISTOGRAM,
                       false)
{
}

int DiagnosticDataRSHistograms::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header;
    PPCNT_RS_Fec_Histograms_Counters::DumpFieldNames(sstream);
    sstream << endl;

    csv_out.WriteBuf(sstream.str());
#endif

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataRSHistograms::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                    IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    struct ppcnt_rs_fec_histograms_counters rs_histograms;
    ppcnt_rs_fec_histograms_counters_unpack(&rs_histograms, (uint8_t *)&(dd.data_set));

    PPCNT_RS_Fec_Histograms_Counters::DumpData(rs_histograms, sstream);
    IBDIAGNET_RETURN_VOID;
}

// PPCNT PLR Counters
DiagnosticDataPLRCounters::DiagnosticDataPLRCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PLR_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PLR_COUNTERS,
                       false)
{
}

int DiagnosticDataPLRCounters::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);

#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    PPCNT_PLR_Counters::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());
#endif

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPLRCounters::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                   IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct ppcnt_plr_counters plr_counters;
    ppcnt_plr_counters_unpack(&plr_counters, (uint8_t *)&(dd.data_set));
    bool has_cap = this->m_p_ibdiag->GetCapabilityModulePtr()->
                    IsSupportedGMPCapability(p_node,
                                             EnGMPCAPIsPLRMaxRetransmissionRateSupported);

    PPCNT_PLR_Counters::DumpData(plr_counters, sstream, has_cap);
    IBDIAGNET_RETURN_VOID;
}


// PPCNT Phy Statistics
DiagnosticDataPhyStatistics::DiagnosticDataPhyStatistics():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_VERSION,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PHY_LAYER_STATISTICS,
                       false,
                       SUPPORT_SW_CA)
{
}

int DiagnosticDataPhyStatistics::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);
#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    PPCNT_Phys_Layer_Stat_Cntrs::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());
#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPhyStatistics::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd,
                                                     IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct phys_layer_stat_cntrs phy_statistics;
    phys_layer_stat_cntrs_unpack(&phy_statistics, (uint8_t *)&(dd.data_set));
    bool is_effective_counter_supported = this->m_p_ibdiag->GetCapabilityModulePtr()->
                          IsSupportedGMPCapability(p_node, EnGMPCAPIsEffectiveCounterSupported);
    bool is_raw_ber_per_lane_supported = this->m_p_ibdiag->GetCapabilityModulePtr()->
            IsSupportedGMPCapability(p_node, EnGMPCAPIsRawBerPerLaneSupported);
    PPCNT_Phys_Layer_Stat_Cntrs::DumpData(phy_statistics, sstream, 
                                          is_effective_counter_supported, 
                                          is_raw_ber_per_lane_supported);
    IBDIAGNET_RETURN_VOID;
}

// PPCNT 0x26 InfiniBand General Counters
DiagnosticDataInfiniBandGeneralCounters::DiagnosticDataInfiniBandGeneralCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_INFINIBAND_GENERAL_COUNTERS,
                       false,
                       SUPPORT_SW_CA)
{
}

int DiagnosticDataInfiniBandGeneralCounters::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);
#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    PPCNT_InfiniBand_General_Counters::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());
#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataInfiniBandGeneralCounters::DumpDiagnosticData(stringstream &sstream,
                                                                 VS_DiagnosticData &dd,
                                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct ppcnt_infiniband_general_counter data;
    ppcnt_infiniband_general_counter_unpack(&data, (uint8_t *)&(dd.data_set));
    PPCNT_InfiniBand_General_Counters::DumpData(data, sstream);
    IBDIAGNET_RETURN_VOID;
}


// PPCNT 0x27 General Counters
DiagnosticDataGeneralCounters::DiagnosticDataGeneralCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_GENERAL_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_GENERAL_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_GENERAL_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_GENERAL_COUNTERS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_GENERAL_COUNTERS,
                       false,
                       SUPPORT_SW)
{
}

int DiagnosticDataGeneralCounters::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);
#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header << ',';
    PPCNT_L1_General_Counters::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());
#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataGeneralCounters::DumpDiagnosticData(stringstream &sstream,
                                                                 VS_DiagnosticData &dd,
                                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct ppcnt_l1_general_counters data;
    ppcnt_l1_general_counters_unpack(&data, (uint8_t *)&(dd.data_set));
    PPCNT_L1_General_Counters::DumpData(data, sstream);
    IBDIAGNET_RETURN_VOID;
}

// PPCNT 0x29 Packet Type Counters
DiagnosticDataPacketTypesCounters::DiagnosticDataPacketTypesCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PACKET_TYPES_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_PACKET_TYPES_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_PACKET_TYPES_COUNTERS_NUM_FIELDS,
                       DIAGNOSTIC_DATA_PACKET_TYPES_COUNTERS_NAME,
                       NSB::get(this),
                       DD_PHY_TYPE,
                       SECTION_PACKET_TYPE_COUNTERS,
                       false,
                       SUPPORT_SW_CA)
{
}

int DiagnosticDataPacketTypesCounters::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    int status = DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(csv_out);
    if (status)
        IBDIAG_RETURN(status);
#ifdef SHOW_PHY_HEADERS
    stringstream sstream;
    sstream << '#' << m_common_header  << ',';
    PPCNT_Packet_Type_Counters::DumpFieldNames(sstream);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());
#endif
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void DiagnosticDataPacketTypesCounters::DumpDiagnosticData(stringstream &sstream,
                                                                 VS_DiagnosticData &dd,
                                                                 IBNode *p_node) const
{
    IBDIAGNET_ENTER;
    struct ppcnt_packet_type_counter data;
    ppcnt_packet_type_counter_unpack(&data, (uint8_t *)&(dd.data_set));
    PPCNT_Packet_Type_Counters::DumpData(data, sstream);
    IBDIAGNET_RETURN_VOID;
}
