/*
 * Copyright (c) 2004-2016 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2022, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"

/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMMadGetSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        u_int64_t am_key,
        u_int8_t class_version,
        const data_func_set_t & attribute_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_AggregationManagement am_mad;
    CLEAR_STRUCT(am_mad);
    //build am lid route header
    this->CommonMadHeaderBuild(&am_mad.MAD_Header_Common,
            IBIS_IB_CLASS_AM,
            method,
            attribute_id,
            attribute_modifier,
            class_version);

    am_mad.am_key = am_key;

    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, sl, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_AM,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_AM,
            data_func_set_t(&am_mad, IBIS_FUNC_LST(MAD_AggregationManagement)),
            attribute_data,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMClassPortInfoGet(u_int16_t lid,
        u_int8_t sl,
        uint64_t am_key,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_ib_class_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,
            am_key,
            1,
            data_func_set_t(p_ib_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMClassPortInfoSet(u_int16_t lid,
        u_int8_t sl,
        uint64_t am_key,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,
            am_key,
            1,
            data_func_set_t(p_ib_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMKeyInfoGet(u_int16_t lid,
        u_int8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_AMKeyInfo *p_am_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_AMKeyInfo Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_KEY_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_am_key_info, IBIS_FUNC_LST(AM_AMKeyInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMKeyInfoSet(u_int16_t lid,
        u_int8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_AMKeyInfo *p_am_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_AMKeyInfo Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_KEY_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_am_key_info, IBIS_FUNC_LST(AM_AMKeyInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMANInfoGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ANInfo *p_an_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_an_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ANInfo Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_AN_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_info, IBIS_FUNC_LST(AM_ANInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMANActiveJobsGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ANActiveJobs *p_an_jobs,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_an_jobs);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ANActiveJobs Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_AN_ACTIVE_JOBS,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_jobs, IBIS_FUNC_LST(AM_ANActiveJobs)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMANInfoSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ANInfo *p_an_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ANInfo Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_AN_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_info, IBIS_FUNC_LST(AM_ANInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMResourceCleanupSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ResourceCleanup_V2 *p_resource_cleanup,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    if (class_version < 2) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR, "Invalid Sending AM_ResourceCleanup_V2 with "
                                     "version numner %u for Set MAD lid = %u\n",
                                     class_version, lid);
        throw std::invalid_argument("Invalid version numnber "
                                    "for sending AM_ResourceCleanup_V2");
    }
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ResourceCleanup_V2 Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_RESOURCE_CLEANUP,
            0,
            am_key,
            class_version,
            data_func_set_t(p_resource_cleanup, IBIS_FUNC_LST(AM_ResourceCleanup_V2)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMResourceCleanupSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ResourceCleanup *p_resource_cleanup,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    if (class_version > 1) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR, "Invalid Sending AM_ResourceCleanup with "
                                     "version numner %u for Set MAD lid = %u\n",
                                     class_version, lid);
        throw std::invalid_argument("Invalid version numnber "
                                    "for sending AM_ResourceCleanup");
    }
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ResourceCleanup Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_RESOURCE_CLEANUP,
            0,
            am_key,
            1,
            data_func_set_t(p_resource_cleanup, IBIS_FUNC_LST(AM_ResourceCleanup)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQPAllocationSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QPAllocation *p_qp_allocation,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AMQPAllocation Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_QP_ALLOCATION,
            0,
            am_key,
            class_version,
            data_func_set_t(p_qp_allocation, IBIS_FUNC_LST(AM_QPAllocation)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQPCConfigGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QPCConfig *p_qpc_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_QPCConfig Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_QPC_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_qpc_config, IBIS_FUNC_LST(AM_QPCConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQPCConfigSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QPCConfig *p_qpc_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_QPCConfig Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_QPC_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_qpc_config, IBIS_FUNC_LST(AM_QPCConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMTreeConfigGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_TreeConfig *p_tree_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_TreeConfig Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_TREE_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_tree_config, IBIS_FUNC_LST(AM_TreeConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMTreeConfigSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_TreeConfig *p_tree_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_TreeConfig Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_TREE_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_tree_config, IBIS_FUNC_LST(AM_TreeConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AM_TreeToJobBindGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_TreeToJobBind *p_tree_job_bind,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_TreeToJobBind Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_TREE_TO_JOB_BIND,
            0,
            am_key,
            class_version,
            data_func_set_t(p_tree_job_bind, IBIS_FUNC_LST(AM_TreeToJobBind)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AM_TreeToJobBindSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_TreeToJobBind *p_tree_job_bind,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_TreeToJobBind Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_TREE_TO_JOB_BIND,
            0,
            am_key,
            class_version,
            data_func_set_t(p_tree_job_bind, IBIS_FUNC_LST(AM_TreeToJobBind)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AM_ANSemaphoreInfoGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ANSemaphoreInfo *p_an_semaphore_info,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_ANSemaphoreInfoGet Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_AN_SEMAPHORE_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_semaphore_info, IBIS_FUNC_LST(AM_ANSemaphoreInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMANSATQPInfoGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_ANSATQPInfo *p_an_sat_qp_info,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AMANSATQPInfoGet Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_AN_SAT_QP_INFO,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_sat_qp_info, IBIS_FUNC_LST(AM_ANSATQPInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQPDatabaseGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QPDatabase *p_an_qp_database,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AMQPDatabaseGet Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_QP_DATABASE,
            0,
            am_key,
            class_version,
            data_func_set_t(p_an_qp_database, IBIS_FUNC_LST(AM_QPDatabase)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQuotaConfigSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QuotaConfig *p_quota_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_QuotaConfig Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_QUOTA_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_quota_config, IBIS_FUNC_LST(AM_QuotaConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMQuotaConfigGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_QuotaConfig *p_quota_config,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_QuotaConfig Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_QUOTA_CONFIG,
            0,
            am_key,
            class_version,
            data_func_set_t(p_quota_config, IBIS_FUNC_LST(AM_QuotaConfig)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMPerformanceCountersGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_PerformanceCounters *p_perf_cntr,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PerformanceCounters Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_PERFORMANCE_COUNTERS,
            0,
            am_key,
            class_version,
            data_func_set_t(p_perf_cntr, IBIS_FUNC_LST(AM_PerformanceCounters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMPerformanceCountersGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        u_int8_t mode,
        struct AM_PerformanceCounters *p_perf_cntr,
        const clbck_data_t *p_clbck_data,
        phys_port_t port_number) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PerformanceCounters Get MAD lid = %u\n", lid);

    u_int32_t attribute_modifier = ((0x03 & mode) << 30) | (u_int32_t)port_number;

    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_PERFORMANCE_COUNTERS,
            attribute_modifier,
            am_key,
            class_version,
            data_func_set_t(p_perf_cntr, IBIS_FUNC_LST(AM_PerformanceCounters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMPerformanceCountersSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        struct AM_PerformanceCounters *p_perf_cntr,
        const clbck_data_t *p_clbck_data) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PerformanceCounters Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_PERFORMANCE_COUNTERS,
            0,
            am_key,
            class_version,
            data_func_set_t(p_perf_cntr, IBIS_FUNC_LST(AM_PerformanceCounters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::AMPerformanceCountersSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        u_int8_t mode,
        struct AM_PerformanceCounters *p_perf_cntr,
        const clbck_data_t *p_clbck_data,
        phys_port_t port_number) {
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PerformanceCounters Set MAD lid = %u\n", lid);

    u_int32_t attribute_modifier = ((0x03 & mode) << 30) | (u_int32_t)port_number;

    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_PERFORMANCE_COUNTERS,
            attribute_modifier,
            am_key,
            class_version,
            data_func_set_t(p_perf_cntr, IBIS_FUNC_LST(AM_PerformanceCounters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMPortCreditResourcesAllocationSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        uint8_t port_number,
        struct AM_PortCreditResourcesAllocation *p_port_credits_res_alloc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PortCreditResourcesAllocation Set MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_PORT_CREDIT_RESOURCES_ALLOCATION,
            port_number,
            am_key,
            class_version,
            data_func_set_t(p_port_credits_res_alloc, IBIS_FUNC_LST(AM_PortCreditResourcesAllocation)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMPortCreditResourcesAllocationGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        uint8_t port_number,
        struct AM_PortCreditResourcesAllocation *p_port_credits_res_alloc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_PortCreditResourcesAllocation Get MAD lid = %u\n", lid);
    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_PORT_CREDIT_RESOURCES_ALLOCATION,
            port_number,
            am_key,
            class_version,
            data_func_set_t(p_port_credits_res_alloc, IBIS_FUNC_LST(AM_PortCreditResourcesAllocation)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMMulticastPrivateLFTSet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        uint16_t mlid_block,
        uint8_t port_group,
        uint8_t plft,
        struct AM_MulticastPrivateLFT *p_multicast_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_MulticastPrivateLFTSet Set MAD lid = %u\n", lid);

    u_int32_t attribute_modifier = ((u_int32_t)port_group << 24) | ((u_int32_t)plft << 16) | mlid_block;

    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_AM_MULTICAST_PRIVATE_LFT,
            attribute_modifier,
            am_key,
            class_version,
            data_func_set_t(p_multicast_table, IBIS_FUNC_LST(AM_MulticastPrivateLFT)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::AMMulticastPrivateLFTGet(u_int16_t lid,
        uint8_t sl,
        uint64_t am_key,
        uint8_t class_version,
        uint16_t mlid_block,
        uint8_t port_group,
        uint8_t plft,
        struct AM_MulticastPrivateLFT *p_multicast_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending AM_MulticastPrivateLFTGet Set MAD lid = %u\n", lid);

    u_int32_t attribute_modifier = ((u_int32_t)port_group << 24) | ((u_int32_t)plft << 16) | mlid_block;

    int rc = this->AMMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_AM_MULTICAST_PRIVATE_LFT,
            attribute_modifier,
            am_key,
            class_version,
            data_func_set_t(p_multicast_table, IBIS_FUNC_LST(AM_MulticastPrivateLFT)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::RegisterAmTrapsHandler(mad_handler_callback_func_t callback_func,
                                 void *context) {

    return RegisterMadHandler(
            IBIS_IB_CLASS_AM,
            IBIS_IB_ATTR_NOTICE,
            IBIS_IB_MAD_METHOD_TRAP,
            IBIS_IB_DATA_OFFSET_AM,
            (const unpack_data_func_t)MAD_AggregationManagement_unpack,
            (const dump_data_func_t)MAD_AggregationManagement_dump,
            (const unpack_data_func_t)Notice_unpack,
            (const dump_data_func_t)Notice_dump,
            callback_func,
            context);
}


int Ibis::RepressAmTrap(ib_address_t *p_ib_address,
                        MAD_AggregationManagement *p_am_mad,
                        Notice *p_notice) {

    p_am_mad->MAD_Header_Common.Method = IBIS_IB_MAD_METHOD_TRAP_REPRESS;
    return MadGetSet(p_ib_address,
            IBIS_IB_CLASS_AM,
            IBIS_IB_MAD_METHOD_TRAP_REPRESS,
            IBIS_IB_DATA_OFFSET_AM,
            data_func_set_t(p_am_mad, IBIS_FUNC_LST(MAD_AggregationManagement)),
            data_func_set_t(p_notice, IBIS_FUNC_LST(Notice)),
            NULL);
}
